from meeko import MoleculePreparation
from meeko import PDBQTWriterLegacy
from rdkit import Chem
from rdkit.Chem import rdDistGeom
import warnings
import pytest


def test1():
    typer = {
        "ATOM_PARAMS": {
            "reduced set": [
                {"smarts": "[#1]", "atype": "H",},
                {"smarts": "[#1][#6]([#7,#8])[#7,#8]", "atype": "HD"}, # one H in oxazole
                {"smarts": "[C]", "atype": "C"},
                {"smarts": "[c]", "atype": "A"},
                {"smarts": "[#7]", "atype": "NA"},
                {"smarts": "[#8]", "atype": "OA"},
            ]
        }
    }
    
    preparator = MoleculePreparation(atom_type_smarts=typer)
    
    mol = Chem.MolFromSmiles("c1ncco1") # oxazole
    mol = Chem.AddHs(mol)
    rdDistGeom.EmbedMolecule(mol)
    setups = preparator.prepare(mol)
    pdbqt_string, is_ok, error_msg = PDBQTWriterLegacy.write_string(setups[0])
    
    count_HD = 0
    for line in pdbqt_string.split("\n"):
        if line.startswith("ATOM") or line.startswith("HETATM"):
            count_HD += int(line[77:79] == "HD")

    assert(count_HD == 1)

def test2():
    typer = {
        "ATOM_PARAMS": {
            "reduced set": [
                {"smarts": "[#1]", "atype": "H_MERGE",},
                {"smarts": "[#1][#6]([#7,#8])[#7,#8]", "atype": "HC"}, # one H in oxazole
                {"smarts": "[#1][#8]", "atype": "HD"},
                {"smarts": "[C]", "atype": "C"},
                {"smarts": "[c]", "atype": "A"},
                {"smarts": "[#7]", "atype": "NA"},
                {"smarts": "[#8]", "atype": "OA"},
            ]
        }
    }
    
    preparator = MoleculePreparation(
        atom_type_smarts=typer,
        merge_these_atom_types=("H", "H_MERGE"),
    )
    
    mol = Chem.MolFromSmiles("c1nc(CO)co1") # oxazole with an -OH
    mol = Chem.AddHs(mol)
    rdDistGeom.EmbedMolecule(mol)
    setups = preparator.prepare(mol)
    pdbqt_string, is_ok, error_msg = PDBQTWriterLegacy.write_string(setups[0])
    
    count_atoms = 0
    count_HD = 0
    count_HC = 0
    for line in pdbqt_string.split("\n"):
        if line.startswith("ATOM") or line.startswith("HETATM"):
            count_atoms += 1
            count_HD += int(line[77:79] == "HD")
            count_HC += int(line[77:79] == "HC")

    assert(count_HD == 1)
    assert(count_HC == 1)
    assert(count_atoms == 9) # 5 (oxazole ring) + 1 (HC type) + 3 (COH) = 9



from meeko import MoleculePreparation
from rdkit import Chem
import pathlib

min_rings_dict = {
    "adamantane": 4,
    "adamantane-stretched": 4,
    "cubane": 6,
    "oxabicyclo-2-2-2-octane": 3,
    "tricycle-6-5-5": 3,
    "tricycle-6-5-6": 3,
    "tricycle-6-5-6_B": 3,
    "tricycle-6-7-7": 2,
    "tricycle-5-6-7": 2,
    "break-equiv": 4,
    "break-equiv-small": 4,
    "break-equiv-b": 4,
    "break-equiv-a": 4,
    "break-equiv-2": 4,
    "adamantanish-5-5-6": 2,
}
workdir = pathlib.Path(__file__)
filenames = {name: str(workdir.parents[0] / "small_cycle_data" / ("%s.sdf" % name)) for name in min_rings_dict}
mk_prep = MoleculePreparation() #keep_equivalent_rings=True)

def run(molname, min_rings):
    filename = filenames[molname]
    mol = Chem.MolFromMolFile(filename, removeHs=False)
    setups = mk_prep.prepare(mol)
    assert(len(setups) == 1)
    setup = setups[0]
    print("\n%s" % molname)
    found_rings = 0
    for ring in setup.rings:
        print("len=%d, atoms:" % (len(ring)), ring)
        found_rings += 1
    print("Got %d, needed %d" % (found_rings, min_rings))
    assert(found_rings >= min_rings)
    

def test_all():
    for molname in min_rings_dict:
        min_rings = min_rings_dict[molname]
        run(molname, min_rings)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#

import argparse
import os
import sys
import json
import warnings

from rdkit import Chem

from meeko import MoleculePreparation
from meeko import rdkitutils
from meeko import PDBQTWriterLegacy

try:
    import prody
    from meeko import CovalentBuilder
    _prody_parsers = {'pdb': prody.parsePDB, 'mmcif': prody.parseMMCIF}
    warnings.warn("Prody not available, covalent docking won't work", ImportWarning)
except:
    _has_prody = False
    _prody_parsers = {}
else:
    _has_prody = True

try:
    from meeko import obutils # fails if openbabel not available
except:
    _has_openbabel = False
else:
    _has_openbabel = True


def cmd_lineparser():
    backend = 'rdkit'
    if '--ob_backend' in sys.argv:
        if not _has_openbabel:
            raise ImportError('--ob_backend requires openbabel which is not available')
        backend = 'ob'
        sys.argv.remove('--ob_backend')
    conf_parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter, add_help=False)
    conf_parser.add_argument('-c', '--config_file',
            help='configure MoleculePreparation from JSON file. Overriden by command line args.')
    confargs, remaining_argv = conf_parser.parse_known_args()

    config = MoleculePreparation.get_defaults_dict()

    if confargs.config_file is not None:
        with open(confargs.config_file) as f:
            c = json.load(f)
            config.update(c)

    parser = argparse.ArgumentParser()#parents=[conf_parser]) # parents shows --config_file in help msg
    parser.set_defaults(**config)
    parser.add_argument("-v", "--verbose", dest="verbose",
                        action="store_true", help="print information about molecule setup")

    io_group = parser.add_argument_group("Input/Output")
    io_group.add_argument("-i", "--mol", dest="input_molecule_filename", required=True,
                        action="store", help="molecule file (MOL2, SDF,...)")
    io_group.add_argument("-o", "--out", dest="output_pdbqt_filename",
                        action="store", help="output pdbqt filename. Single molecule input only.")
    io_group.add_argument("--multimol_outdir", dest="multimol_output_dir",
                        action="store", help="folder to write output pdbqt for multi-mol inputs. Incompatible with -o/--out and -/--.")
    io_group.add_argument("--multimol_prefix", dest="multimol_prefix",
                        action="store", help="replace internal molecule name in multi-molecule input by specified prefix. Incompatible with -o/--out and -/--.")
    io_group.add_argument('-', '--',  dest='redirect_stdout', action='store_true',
                        help='do not write file, redirect output to STDOUT. Argument -o/--out is ignored. Single molecule input only.')

    config_group = parser.add_argument_group("Molecule preparation")
    config_group.add_argument('-c', '--config_file',
            help='configure MoleculePreparation from JSON file. Overriden by command line args.') # parsed above by conf_parser, here for help msg
    config_group.add_argument("--rigid_macrocycles",dest="rigid_macrocycles",
                        action="store_true", help="keep macrocycles rigid in input conformation")
    config_group.add_argument("--keep_chorded_rings",dest="keep_chorded_rings",
                        action="store_true", help="return all rings from exhaustive perception")
    config_group.add_argument("--keep_equivalent_rings",dest="keep_equivalent_rings",
                        action="store_true", help="equivalent rings have the same size and neighbors")
    config_group.add_argument("--min_ring_size", dest="min_ring_size",
                        type=int, help="min nr of atoms in ring for opening")
    config_group.add_argument("-w", "--hydrate", dest="hydrate",
                        action="store_true", help="add water molecules for hydrated docking")
    config_group.add_argument("--merge_these_atom_types", dest="merge_these_atom_types", nargs="*",
                        help="list of atom types to merge, default is \"H\"", default=["H"])
    config_group.add_argument("-r", "--rigidify_bonds_smarts", dest="rigidify_bonds_smarts",
                        action="append", help="SMARTS patterns to rigidify bonds",
                        metavar='SMARTS')
    config_group.add_argument("-b", "--rigidify_bonds_indices", dest="rigidify_bonds_indices",
                        action="append", help="indices of two atoms (in the SMARTS) that define a bond (start at 1)",
                        nargs='+', type=int, metavar='i j')
    config_group.add_argument("-a", "--flexible_amides", dest="flexible_amides",
                        action="store_true", help="allow amide bonds to rotate and be non-planar, which is bad")
    config_group.add_argument("-p", "--atom_type_smarts", dest="atom_type_smarts_json",
                        action="store", help="SMARTS based atom typing (JSON format)")
    config_group.add_argument("-aa", "--add_atom_types", dest="add_atom_types_json",
                        action="append", help="Additional atom types to assign (JSON formated)", metavar="[{'smarts': '<smarts pattern>', 'atype': ',atomtype name>'}, {'smarts': '<smarts pattern>', 'atype': ',atomtype name>'}]")
    config_group.add_argument("--double_bond_penalty", help="penalty > 100 prevents breaking double bonds", type=int)
    config_group.add_argument("--bad_charge_ok", help="NaN and Inf charges allowed in PDBQT", action="store_true")
    config_group.add_argument("--add_index_map", dest="add_index_map",
                        action="store_true", help="write map of atom indices from input to pdbqt")
    config_group.add_argument("--remove_smiles", dest="remove_smiles",
                        action="store_true", help="do not write smiles as remark to pdbqt")
    reactive_group = parser.add_argument_group("Reactive docking")
    reactive_group.add_argument('--reactive_smarts',  help="SMARTS pattern for reactive group")
    reactive_group.add_argument('--reactive_smarts_idx', help='index (1-based) of the reactive atom in --reactive_smarts', type=int)

    need_prody_msg = ''
    if not _has_prody: need_prody_msg = ". Needs Prody which is unavailable"
    covalent_group = parser.add_argument_group("Covalent docking (tethered)%s" % (need_prody_msg))
    covalent_group.add_argument('--receptor', help='receptor filename. Supported formats: [%s]%s' % (
        '/'.join(list(_prody_parsers.keys())),
        need_prody_msg))
    covalent_group.add_argument('--rec_residue',
                                help='examples: "A:LYS:204", "A:HIS:", ":LYS:"')
    covalent_group.add_argument('--tether_smarts',
                                help='SMARTS pattern to define ligand atoms for receptor attachment')
    covalent_group.add_argument('--tether_smarts_indices', type=int, nargs=2, required=False,
                                metavar='IDX', default=[1, 2],
                                help='indices (1-based) of the SMARTS atoms that will be attached (default: 1 2)')

    args = parser.parse_args(remaining_argv)

    # check reactive arguments
    if (args.reactive_smarts is None) != (args.reactive_smarts_idx is None):
        print("Arguments --reactive_smarts and --reactive_smarts_idx require each other", file=sys.stderr)
        sys.exit(2)
    elif args.reactive_smarts_idx is not None:
        if args.reactive_smarts_idx < 1:
            print("--reactive_smarts_idx is 1-indexed, but got %d" % args.reactive_smarts_idx, file=sys.stderr)
            sys.exit(2)
        args.reactive_smarts_idx -= 1 # convert from 1- to 0-index

    # command line arguments override config
    for key in config:
        if key in args.__dict__:
            config[key] = args.__dict__[key]

    if args.atom_type_smarts_json is not None:
        with open(args.atom_type_smarts_json) as f:
            config['atom_type_smarts'] = json.load(f)
    
    if args.add_atom_types_json is not None:
        additional_ats = []
        for at in args.add_atom_types_json:
            at = json.loads(at)
            if type(at) == dict:
                additional_ats.append(at)
            elif type(at) == list:
                additional_ats.extend(at)
        config['add_atom_types'] = additional_ats

    if args.multimol_output_dir is not None or args.multimol_prefix is not None:
        if args.output_pdbqt_filename is not None:
            print("Warning: -o/--out ignored with --multimol_outdir or --multimol_prefix", file=sys.stderr)
        if args.redirect_stdout:
            print("Warning: -/-- ignored with --multimol_outdir or --multimol_prefix", file=sys.stderr)

    # verify sanity of covalent docking input
    num_required_covalent_args = 0
    num_required_covalent_args += int(args.receptor is not None)
    num_required_covalent_args += int(args.rec_residue is not None)
    num_required_covalent_args += int(args.tether_smarts is not None)
    if num_required_covalent_args not in [0, 3]:
        print("Error: --receptor, --rec_residue, and --tether_smarts are all required for covalent docking.")
        sys.exit(2)
    is_covalent = num_required_covalent_args == 3
    if is_covalent and not _has_prody:
        raise ImportError("Covalent docking requires Prody which is not available")
    if min(args.tether_smarts_indices) < 1:
        print("--tether_smarts_indices is 1-indexed, all values must be greater than zero", file=sys.stderr) 
        sys.exit(2)
    args.tether_smarts_indices = [i-1 for i in args.tether_smarts_indices] # convert to 0-index

    # verify sanity of SMARTS patterns to make bonds rigid and convert to 0-based indices
    rigidify_bonds_smarts = config['rigidify_bonds_smarts']
    rigidify_bonds_indices = config['rigidify_bonds_indices']
    if len(rigidify_bonds_indices) != len(rigidify_bonds_smarts):
        raise RuntimeError('length of --rigidify_bonds_indices differs from length of --rigidify_bonds_smarts')
    for indices in rigidify_bonds_indices:
        if len(indices) != 2:
            raise RuntimeError('--rigidify_bonds_indices must specify pairs, e.g. -b 1 2 -b 3 4')
        indices[0] = indices[0] - 1 # convert from 1- to 0-index
        indices[1] = indices[1] - 1

    return args, config, backend, is_covalent


class Output:
    def __init__(self, multimol_output_dir, multimol_prefix, redirect_stdout, output_filename):
        is_multimol = (multimol_prefix is not None) or (multimol_output_dir is not None)
        self._mkdir(multimol_output_dir)

        if multimol_output_dir is None:
            multimol_output_dir = '.'
        self.multimol_output_dir = multimol_output_dir
        self.multimol_prefix = multimol_prefix
        self.redirect_stdout = redirect_stdout
        self.output_filename = output_filename
        self.is_multimol = is_multimol
        self.visited_filenames = set()
        self.duplicate_filenames = set()
        self.visited_names = set()
        self.duplicate_names = set()
        self.num_files_written = 0
        self.counter = 0

    def __call__(self, pdbqt_string, name, suffixes=()):
        self.counter += 1
        if name in self.visited_names:
            self.duplicate_names.add(name)
        self.visited_names.add(name)
        if self.multimol_prefix is not None:
            name = '%s-%d' % (args.multimol_prefix, self.counter)
        for suffix in suffixes:
            if suffix is not None and len(suffix) > 0:
                name += "_" + suffix
        if self.is_multimol:
            if name in self.visited_filenames:
                self.duplicate_filenames.add(name)
                repeat_id = 1
                newname = name + "-again%d" % repeat_id
                while newname in self.visited_filenames:
                    repeat_id += 1
                    newname = name + "-again%d" % repeat_id
                print("Renaming %s to %s to disambiguate filename" % (name, newname), file=sys.stderr)
                name = newname
            self.visited_filenames.add(name)
            fpath = os.path.join(self.multimol_output_dir, name + '.pdbqt')
            print(pdbqt_string, end='', file=open(fpath, 'w'))
            self.num_files_written += 1
        elif self.redirect_stdout:
            print(pdbqt_string, end='')
        else:
            if self.output_filename is None:
                filename = '%s.pdbqt' % name
            else:
                filename = self.output_filename
            print(pdbqt_string, end='', file=open(filename, 'w'))
            self.num_files_written += 1

    def _mkdir(self, multimol_output_dir):
        """make directory if it doesn't exist yet """
        if multimol_output_dir is not None:
            if not os.path.exists(multimol_output_dir):
                os.mkdir(multimol_output_dir)

    def get_duplicates_info_string(self):
        if not self.is_multimol:
            return None
        if len(self.duplicate_names):
            # with multimol_prefix with can have duplicate molecule names,
            # but not duplicate filenames. This warning is for such cases.
            string = "Warning: %d molecules have duplicated names" % len(self.duplicate_names)
        else:
            string = "No duplicate molecule molecule names were found"
        # if we have duplicate_filenames, we also have duplicate molecule names,
        # but it suffices to return the string below
        if len(self.duplicate_filenames):
            string = "Warning: %d molecules with repeated names were suffixed with -again<n>" % len(self.duplicate_filenames)
        return string

    @staticmethod
    def get_suffixes(molsetups):
        if len(molsetups) == 1:
            return ("",) # no suffix needed
        else:
            return tuple("mk%d" % (i + 1) for i in range(len(molsetups)))

def main():
    args, config, backend, is_covalent = cmd_lineparser()
    input_molecule_filename = args.input_molecule_filename

    # read input
    input_fname, ext = os.path.splitext(input_molecule_filename)
    ext = ext[1:].lower()
    if backend == 'rdkit':
        parsers = {'sdf': Chem.SDMolSupplier, 'mol2': rdkitutils.Mol2MolSupplier, 'mol': Chem.SDMolSupplier}
        if not ext in parsers:
            print("*ERROR* Format [%s] not in supported formats [%s]" % (ext, '/'.join(list(parsers.keys()))))
            sys.exit(1)
        mol_supplier = parsers[ext](input_molecule_filename, removeHs=False) # input must have explicit H
    elif backend == 'ob':
        print("Using openbabel instead of rdkit")
        mol_supplier = obutils.OBMolSupplier(input_molecule_filename, ext)

    # configure output writer
    if args.output_pdbqt_filename is None:
        output_filename = input_fname + '.pdbqt'
    else:
        output_filename = args.output_pdbqt_filename
    output = Output(
            args.multimol_output_dir,
            args.multimol_prefix,
            args.redirect_stdout,
            output_filename,
            )

    # initialize covalent object for receptor
    if is_covalent:
        rec_filename = args.receptor
        _, rec_extension = os.path.splitext(rec_filename)
        rec_extension = rec_extension[1:].lower()
        prody_parser = _prody_parsers[rec_extension]
        rec_prody_mol = prody_parser(rec_filename)
        covalent_builder = CovalentBuilder(rec_prody_mol, args.rec_residue)

    input_mol_counter = 0
    input_mol_skipped = 0
    input_mol_with_failure = 0 # if reactive or covalent, each mol can yield multiple PDBQT
    nr_failures = 0
    is_after_first = False
    preparator = MoleculePreparation.from_config(config)
    for mol in mol_supplier:
        if is_after_first and output.is_multimol == False:
            print("Processed only the first molecule of multiple molecule input.")
            print("Use --multimol_prefix and/or --multimol_outdir to process all molecules in %s." % (
                input_molecule_filename))
            break
        is_after_first = True

        # check that molecule was successfully loaded
        if backend == 'rdkit':
            is_valid = mol is not None
        elif backend == 'ob':
            is_valid = mol.NumAtoms() > 0
        input_mol_skipped += int(is_valid==False)
        if not is_valid: continue

        this_mol_had_failure = False

        if is_covalent:
            for cov_lig in covalent_builder.process(mol, args.tether_smarts, args.tether_smarts_indices):
                root_atom_index = cov_lig.indices[0]
                molsetups = preparator.prepare(cov_lig.mol, root_atom_index=root_atom_index, not_terminal_atoms=[root_atom_index])
                res, chain, num = cov_lig.res_id
                suffixes = output.get_suffixes(molsetups)
                for molsetup, suffix in zip(molsetups, suffixes):
                    pdbqt_string, success, error_msg = PDBQTWriterLegacy.write_string(molsetup, bad_charge_ok=args.bad_charge_ok)
                    if success:
                        pdbqt_string = PDBQTWriterLegacy.adapt_pdbqt_for_autodock4_flexres(pdbqt_string, res, chain, num)
                        name = molsetup.name
                        output(pdbqt_string, name, (cov_lig.label, suffix))
                    else:
                        nr_failures += 1
                        this_mol_had_failure = True
                        print(error_msg, file=sys.stderr)
                        
        else:
            molsetups = preparator.prepare(mol)
            suffixes = output.get_suffixes(molsetups) 
            for molsetup, suffix in zip(molsetups, suffixes):
                pdbqt_string, success, error_msg = PDBQTWriterLegacy.write_string(molsetup, bad_charge_ok=args.bad_charge_ok)
                if success:
                    name = molsetup.name
                    output(pdbqt_string, name, (suffix,))
                else:
                    nr_failures += 1
                    this_mol_had_failure = True
                    print(error_msg, file=sys.stderr)
                    
            if args.verbose: preparator.show_setup()

        input_mol_with_failure += int(this_mol_had_failure)

    if output.is_multimol:
        print("Input molecules processed: %d, skipped: %d" % (output.counter, input_mol_skipped))
        print("PDBQT files written: %d" % (output.num_files_written))
        print("PDBQT files not written due to error: %d" % (nr_failures))
        print("Input molecules with errors: %d" % (input_mol_with_failure))
        print(output.get_duplicates_info_string())
    return

if __name__ == '__main__':
    sys.exit(main())
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko preparation
#

from inspect import signature
import os
import sys
from collections import OrderedDict
import warnings

from rdkit import Chem

from .molsetup import OBMoleculeSetup
from .molsetup import RDKitMoleculeSetup
from .atomtyper import AtomTyper
from .bondtyper import BondTyperLegacy
from .hydrate import HydrateMoleculeLegacy
from .macrocycle import FlexMacrocycle
from .flexibility import FlexibilityBuilder
from .writer import PDBQTWriterLegacy
from .reactive import assign_reactive_types

try:
    from openbabel import openbabel as ob
except ImportError:
    _has_openbabel = False
else:
    _has_openbabel = True

# DeprecationWarning is not displayed by default
warnings.filterwarnings("default", category=DeprecationWarning)

class MoleculePreparation:
    def __init__(self,
            merge_these_atom_types=("H",),
            hydrate=False,
            flexible_amides=False,
            rigid_macrocycles=False,
            min_ring_size=7,
            max_ring_size=33,
            keep_chorded_rings=False,
            keep_equivalent_rings=False,
            double_bond_penalty=50,
            rigidify_bonds_smarts=[],
            rigidify_bonds_indices=[],
            atom_type_smarts={},
            add_atom_types=[],
            reactive_smarts=None,
            reactive_smarts_idx=None,
            add_index_map=False,
            remove_smiles=False,
        ):

        self.deprecated_setup_access = None
        self.merge_these_atom_types = merge_these_atom_types
        self.hydrate = hydrate
        self.flexible_amides = flexible_amides
        self.rigid_macrocycles = rigid_macrocycles
        self.min_ring_size = min_ring_size
        self.max_ring_size = max_ring_size
        self.keep_chorded_rings = keep_chorded_rings
        self.keep_equivalent_rings = keep_equivalent_rings
        self.double_bond_penalty = double_bond_penalty
        self.rigidify_bonds_smarts = rigidify_bonds_smarts
        self.rigidify_bonds_indices = rigidify_bonds_indices
        self.atom_type_smarts = atom_type_smarts
        self.add_atom_types = add_atom_types
        self.reactive_smarts = reactive_smarts
        self.reactive_smarts_idx = reactive_smarts_idx
        self.add_index_map = add_index_map
        self.remove_smiles = remove_smiles

        self._atom_typer = AtomTyper(self.atom_type_smarts, self.add_atom_types)
        self._bond_typer = BondTyperLegacy()
        self._macrocycle_typer = FlexMacrocycle(
                self.min_ring_size, self.max_ring_size, self.double_bond_penalty)
        self._flex_builder = FlexibilityBuilder()
        self._water_builder = HydrateMoleculeLegacy()
        self._classes_setup = {Chem.rdchem.Mol: RDKitMoleculeSetup}
        if _has_openbabel:
            self._classes_setup[ob.OBMol] = OBMoleculeSetup
        if keep_chorded_rings and keep_equivalent_rings==False:
            warnings.warn("keep_equivalent_rings=False ignored because keep_chorded_rings=True", RuntimeWarning)
        if (reactive_smarts is None) != (reactive_smarts_idx is None):
            raise ValueError("reactive_smarts and reactive_smarts_idx require each other")

    @property
    def setup(self):
        msg = "MoleculePreparation.setup is deprecated in Meeko v0.5."
        msg += " MoleculePreparation.prepare() returns a list of MoleculeSetup instances."
        warnings.warn(msg, DeprecationWarning)
        return self.deprecated_setup_access

    @classmethod
    def get_defaults_dict(cls):
        defaults = {}
        sig = signature(cls)
        for key in sig.parameters:
            defaults[key] = sig.parameters[key].default 
        return defaults

    @ classmethod
    def from_config(cls, config):
        expected_keys = cls.get_defaults_dict().keys()
        bad_keys = [k for k in config if k not in expected_keys]
        for key in bad_keys:
            print("ERROR: unexpected key \"%s\" in MoleculePreparation.from_config()" % key, file=sys.stderr)
        if len(bad_keys) > 0:
            raise ValueError
        p = cls(**config)
        return p

    def prepare(self,
            mol,
            root_atom_index=None,
            not_terminal_atoms=[],
            delete_ring_bonds=[],
            glue_pseudo_atoms={},
            conformer_id=-1,
        ):
        """ 
        Create molecule setup from RDKit molecule

        Args:
            mol (rdkit.Chem.rdchem.Mol): with explicit hydrogens and 3D coordinates
            root_atom_index (int): to set ROOT of torsion tree instead of searching
            not_terminal_atoms (list): make bonds with terminal atoms rotatable
                                       (e.g. C-Alpha carbon in flexres)
            delete_ring_bonds (list): bonds deleted for macrocycle flexibility
                                      each bond is a tuple of two ints (atom 0-indices)
            glue_pseudo_atoms (dict): keys are parent atom indices, values are (x, y, z)
        """
        mol_type = type(mol)
        if not mol_type in self._classes_setup:
            raise TypeError("Molecule is not an instance of supported types: %s" % type(mol))
        setup_class = self._classes_setup[mol_type]
        setup = setup_class.from_mol(mol,
            keep_chorded_rings=self.keep_chorded_rings,
            keep_equivalent_rings=self.keep_equivalent_rings,
            conformer_id=conformer_id,
            )

        self.check_external_ring_break(setup, delete_ring_bonds, glue_pseudo_atoms)

        # 1.  assign atom types (including HB types, vectors and stuff)
        # DISABLED TODO self.atom_typer.set_parm(mol)
        self._atom_typer(setup)
        # 2a. add pi-model + merge_h_pi (THIS CHANGE SOME ATOM TYPES)
        # disabled

        # merge hydrogens (or any terminal atoms)
        indices = set()
        for atype_to_merge in self.merge_these_atom_types:
            for index, atype in setup.atom_type.items():
                if atype == atype_to_merge:
                    indices.add(index)
        setup.merge_terminal_atoms(indices)

        # 3.  assign bond types by using SMARTS...
        #     - bonds should be typed even in rings (but set as non-rotatable)
        #     - if macrocycle is selected, they will be enabled (so they must be typed already!)
        self._bond_typer(setup, self.flexible_amides, self.rigidify_bonds_smarts, self.rigidify_bonds_indices, not_terminal_atoms)
        # 4 . hydrate molecule
        if self.hydrate:
            self._water_builder.hydrate(setup)
        # 5.  break macrocycles into open/linear form
        if self.rigid_macrocycles:
            break_combo_data = None
            bonds_in_rigid_rings = None # not true, but this is only needed when breaking macrocycles
        else:
            break_combo_data, bonds_in_rigid_rings = self._macrocycle_typer.search_macrocycle(setup, delete_ring_bonds)

        # 6.  build flexibility...
        # 6.1 if macrocycles typed:
        #     - walk the setup graph by skipping proposed closures
        #       and score resulting flex_trees basing on the lenght
        #       of the branches generated
        #     - actually break the best closure bond (THIS CHANGES SOME ATOM TYPES)
        # 6.2  - walk the graph and build the flextree
        # 7.  but disable all bonds that are in rings and not
        #     in flexible macrocycles
        # TODO restore legacy AD types for PDBQT
        #self._atom_typer.set_param_legacy(mol)

        setup = self._flex_builder(setup,
                                   root_atom_index=root_atom_index,
                                   break_combo_data=break_combo_data,
                                   bonds_in_rigid_rings=bonds_in_rigid_rings,
                                   glue_pseudo_atoms=glue_pseudo_atoms,
        )

        if self.reactive_smarts is None:
            setups = [setup]
        else:
            reactive_types_dicts = assign_reactive_types(
                    setup,
                    self.reactive_smarts,
                    self.reactive_smarts_idx,
            )
            setups = []
            for r in reactive_types_dicts:
                new_setup = setup.copy()
                new_setup.atom_type = r
                setups.append(new_setup)

        self.deprecated_setup_access = setups[0] # for a gentle introduction of the new API
        return setups


    @staticmethod
    def check_external_ring_break(molsetup, break_ring_bonds, glue_pseudo_atoms):
        for (index1, index2) in break_ring_bonds:
            has_bond = molsetup.get_bond_id(index1, index2) in molsetup.bond
            if not has_bond:
                raise ValueError("bond (%d, %d) not in molsetup" % (index1, index2))
            for index in (index1, index2):
                if index not in glue_pseudo_atoms:
                    raise ValueError("missing glue pseudo for atom %d" % index) 
                xyz = glue_pseudo_atoms[index]
                if len(xyz) != 3:
                    raise ValueError("expected 3 coordinates (got %d) for glue pseudo of atom %d" % (len(xyz), index)) 


    def write_pdbqt_string(self, add_index_map=None, remove_smiles=None):
        msg = "MoleculePreparation.write_pdbqt_string() is deprecated in Meeko v0.5."
        msg += " Pass the MoleculeSetup instance to PDBQTWriterLegacy.write_string()."
        msg += " MoleculePreparation.prepare() returns a list of MoleculeSetup instances."
        warnings.warn(msg, DeprecationWarning)
        pdbqt_string, is_ok, err_msg = PDBQTWriterLegacy.write_string(self.setup)
        if not is_ok:
            msg = 'Cannot generate PDBQT, error from PDBQTWriterLegacy:' + os.linesep
            msg += err_msg
            raise RuntimeError(msg)
        return pdbqt_string


    def write_pdbqt_file(self, pdbqt_filename, add_index_map=None, remove_smiles=None):
        warnings.warn("MoleculePreparation.write_pdbqt_file() is deprecated since Meeko v0.5", DeprecationWarning)
        with open(pdbqt_filename,'w') as w:
            w.write(self.write_pdbqt_string(add_index_map, remove_smiles))




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko
#

from collections import defaultdict
import json
from os import linesep as os_linesep
import pathlib
import sys

import numpy as np
from scipy import spatial

from .utils.covalent_radius_table import covalent_radius
from .utils.autodock4_atom_types_elements import autodock4_atom_types_elements
from .reactive import get_reactive_atype

pkg_dir = pathlib.Path(__file__).parents[0]
with open(pkg_dir / "data" / "residue_params.json") as f:
    residue_params = json.load(f)
with open(pkg_dir / "data" / "flexres_templates.json") as f:
    flexres_templates = json.load(f)
# the above is controversial, see
# https://stackoverflow.com/questions/6028000/how-to-read-a-static-file-from-inside-a-python-package

def _write_pdbqt_line(atomidx, x, y, z, charge, atom_name, res_name, res_num, atom_type, chain,
                      alt_id=" ", in_code="", occupancy=1.0, temp_factor=0.0, record_type="ATOM"):
    if len(atom_name) > 4:
        raise ValueError("max length of atom_name is 4 but atom name is %s" % atom_name)
    atom_name = "%-3s" % atom_name
    line = "{:6s}{:5d} {:4s}{:1s}{:3s} {:1s}{:4d}{:1s}   {:8.3f}{:8.3f}{:8.3f}{:6.2f}{:6.2f}    {:6.3f} {:<2s}"
    line += os_linesep
    return line.format(record_type, atomidx, atom_name, alt_id, res_name, chain,
                   res_num, in_code, x, y, z,
                   occupancy, temp_factor, charge, atom_type)


atom_property_definitions = {'H': 'vdw', 'C': 'vdw', 'A': 'vdw', 'N': 'vdw', 'P': 'vdw', 'S': 'vdw',
                             'Br': 'vdw', 'I': 'vdw', 'F': 'vdw', 'Cl': 'vdw',
                             'NA': 'hb_acc', 'OA': 'hb_acc', 'SA': 'hb_acc', 'OS': 'hb_acc', 'NS': 'hb_acc',
                             'HD': 'hb_don', 'HS': 'hb_don',
                             'Mg': 'metal', 'Ca': 'metal', 'Fe': 'metal', 'Zn': 'metal', 'Mn': 'metal',
                             'MG': 'metal', 'CA': 'metal', 'FE': 'metal', 'ZN': 'metal', 'MN': 'metal',
                             'W': 'water',
                             'G0': 'glue', 'G1': 'glue', 'G2': 'glue', 'G3': 'glue',
                             'CG0': 'glue', 'CG1': 'glue', 'CG2': 'glue', 'CG3': 'glue'}


def _read_receptor_pdbqt_string(pdbqt_string, skip_typing=False):
    atoms = []
    atoms_dtype = [('idx', 'i4'), ('serial', 'i4'), ('name', 'U4'), ('resid', 'i4'),
                   ('resname', 'U3'), ('chain', 'U1'), ("xyz", "f4", (3)),
                   ('partial_charges', 'f4'), ('atom_type', 'U2'),
                   ('alt_id', 'U1'), ('in_code', 'U1'),
                   ('occupancy', 'f4'), ('temp_factor', 'f4'), ('record_type', 'U6')
                  ]
    atom_annotations = {'hb_acc': [], 'hb_don': [],
                        'all': [], 'vdw': [],
                        'metal': []}
    # TZ is a pseudo atom for AutoDock4Zn FF
    pseudo_atom_types = ['TZ']

    idx = 0
    for line in pdbqt_string.split(os_linesep):
        if line.startswith('ATOM') or line.startswith("HETATM"):
            serial = int(line[6:11].strip())
            name = line[12:16].strip()
            resname = line[17:20].strip()
            chainid = line[21].strip()
            resid = int(line[22:26].strip())
            xyz = np.array([line[30:38].strip(), line[38:46].strip(), line[46:54].strip()], dtype=np.float32)
            try:
                partial_charges = float(line[71:77].strip())
            except:
                partial_charges = None # probably reading a PDB, not PDBQT
            atom_type = line[77:79].strip()
            alt_id = line[16:17].strip()
            in_code = line[26:27].strip()
            try:
                occupancy = float(line[54:60])
            except:
                occupancy = None
            try:
                temp_factor = float(line[60:68])
            except:
                temp_factor = None
            record_type = line[0:6].strip()

            if skip_typing:
                atoms.append((idx, serial, name, resid, resname, chainid, xyz, partial_charges, atom_type,
                              alt_id, in_code, occupancy, temp_factor, record_type))
                continue
            if not atom_type in pseudo_atom_types:
                atom_annotations['all'].append(idx)
                atom_annotations[atom_property_definitions[atom_type]].append(idx)
                atoms.append((idx, serial, name, resid, resname, chainid, xyz, partial_charges, atom_type,
                              alt_id, in_code, occupancy, temp_factor, record_type))

            idx += 1

    atoms = np.array(atoms, dtype=atoms_dtype)

    return atoms, atom_annotations


def _identify_bonds(atom_idx, positions, atom_types):
    bonds = defaultdict(list)
    KDTree = spatial.cKDTree(positions)
    bond_allowance_factor = 1.1
    # If we ask more than the number of coordinates/element
    # in the BHTree, we will end up with some inf values
    k = 5 if len(atom_idx) > 5 else len(atom_idx)
    atom_idx = np.array(atom_idx)

    for atom_i, position, atom_type in zip(atom_idx, positions, atom_types):
        distances, indices = KDTree.query(position, k=k)
        r_cov = covalent_radius[autodock4_atom_types_elements[atom_type]]

        optimal_distances = [bond_allowance_factor * (r_cov + covalent_radius[autodock4_atom_types_elements[atom_types[i]]]) for i in indices[1:]]
        bonds[atom_i] = atom_idx[indices[1:][np.where(distances[1:] < optimal_distances)]].tolist()

    return bonds


class PDBQTReceptor:

    flexres_templates = flexres_templates
    skip_types=("H",)

    def __init__(self, pdbqt_filename, skip_typing=False):
        self._pdbqt_filename = pdbqt_filename
        self._atoms = None
        self._atom_annotations = None
        self._KDTree = None

        with open(pdbqt_filename) as f:
            pdbqt_string = f.read()

        self._atoms, self._atom_annotations = _read_receptor_pdbqt_string(pdbqt_string, skip_typing)
        # We add to the KDTree only the rigid part of the receptor
        self._KDTree = spatial.cKDTree(self._atoms['xyz'])
        self._bonds = _identify_bonds(self._atom_annotations['all'], self._atoms['xyz'], self._atoms['atom_type'])
        self.atom_idxs_by_res = self.get_atom_indices_by_residue(self._atoms)

    def __repr__(self):
        return ('<Receptor from PDBQT file %s containing %d atoms>' % (self._pdbqt_filename, self._atoms.shape[0]))

    @staticmethod
    def get_atom_indices_by_residue(atoms):
        """ return a dictionary where residues are keys and
             values are lists of atom indices

            >>> atom_idx_by_res = {("A", "LYS", 417): [0, 1, 2, 3, ..., 8]}
        """

        atom_idx_by_res = {}
        for atom_index, atom in enumerate(atoms):
            res_id = (atom["chain"], atom["resname"], atom["resid"])
            atom_idx_by_res.setdefault(res_id, [])
            atom_idx_by_res[res_id].append(atom_index)
        return atom_idx_by_res

    @staticmethod
    def get_params_for_residue(resname, atom_names, residue_params=residue_params):
        excluded_params = ("atom_names", "bond_cut_atoms", "bonds")
        atom_params = {}
        atom_counter = 0
        err = ""
        ok = True
        is_matched = False
        for terminus in ["", "N", "C"]: # e.g. "CTYR" for C-term TYR, hard-coded in residue_params
            r_id = "%s%s" % (terminus, resname)
            if r_id not in residue_params:
                err = "residue %s not in residue_params" % r_id + os_linesep
                ok = False
                return atom_params, ok, err
            ref_names = set(residue_params[r_id]["atom_names"])
            query_names = set(atom_names)
            if ref_names == query_names:
                is_matched = True
                break

        if not is_matched:
            ok = False
            return atom_params, ok, err

        for atom_name in atom_names:
            name_index = residue_params[r_id]["atom_names"].index(atom_name)
            for param in residue_params[r_id].keys():
                if param in excluded_params:
                    continue
                if param not in atom_params:
                    atom_params[param] = [None] * atom_counter
                value = residue_params[r_id][param][name_index]
                atom_params[param].append(value)
            atom_counter += 1

        return atom_params, ok, err

    def assign_types_charges(self, residue_params=residue_params):
        wanted_params = ("atom_types", "gasteiger")
        atom_params = {key: [] for key in wanted_params}
        ok = True
        err = ""
        for r_id, atom_indices in self.atom_idxs_by_res.items():
            atom_names = tuple(self.atoms(atom_indices)["name"])
            resname = r_id[1]
            params_this_res, ok_, err_ = self.get_params_for_residue(resname, atom_names, residue_params)
            ok &= ok_
            err += err_
            if not ok_:
                print("did not match %s with template" % str(r_id), file=sys.stderr)
                continue
            for key in wanted_params:
                atom_params[key].extend(params_this_res[key])
        if ok:
            self._atoms["partial_charges"] = atom_params["gasteiger"]
            self._atoms["atom_type"] = atom_params["atom_types"]
        return ok, err

    def write_flexres_from_template(self, res_id, atom_index=0):
        success = True
        error_msg = ""
        branch_offset = atom_index # templates assume first atom is 1
        output = {"pdbqt": "", "flex_indices": [], "atom_index": atom_index}
        resname = res_id[1]
        if resname not in self.flexres_templates:
            success = False
            error_msg = "no flexible residue template for resname %s, sorry" % resname
            return output, success, error_msg
        if res_id not in self.atom_idxs_by_res:
            success = False
            chains = set(self._atoms["chain"])
            error_msg += "could not find residue with chain='%s', resname=%s, resnum=%d" % res_id + os_linesep
            error_msg += "chains in this receptor: %s" % ", ".join("'%s'" % c for c in chains) + os_linesep
            if " " in chains: # should not happen because we use strip() when parsing the chain
                error_msg += "use ' ' (a space character) for empty chain" + os_linesep
            if "" in chains:
                error_msg += "use '' (empty string) for empty chain" + os_linesep
            return output, success, error_msg

        # collect lines of res_id
        atoms_by_name = {}
        for i in self.atom_idxs_by_res[res_id]:
            name = self._atoms[i]["name"]
            if name in ['C', 'N', 'O', 'H', 'H1', 'H2', 'H3', 'OXT']: # skip backbone atoms
                continue
            atype = self._atoms[i]["atom_type"]
            if atype in self.skip_types:
                continue
            output["flex_indices"].append(i)
            atoms_by_name[name] = self.atoms(i)

        # check it was a full match
        template = self.flexres_templates[resname]
        got_atoms = set(atoms_by_name)
        ref_atoms = set()
        for i in range(len(template["is_atom"])):
            if template["is_atom"][i]:
                ref_atoms.add(template["atom_name"][i])
        if got_atoms != ref_atoms:
            success = False
            error_msg += "mismatch in atom names for residue %s" % str(res_id) + os_linesep
            error_msg += "names found but not in template: %s" % str(got_atoms.difference(ref_atoms)) + os_linesep
            error_msg += "missing names: %s" % str(ref_atoms.difference(got_atoms)) + os_linesep
            return output, success, error_msg

        # create output string
        n_lines = len(template['is_atom'])
        for i in range(n_lines):
            if template['is_atom'][i]:
                atom_index += 1
                name = template['atom_name'][i]
                atom = atoms_by_name[name]
                if atom["atom_type"] not in self.skip_types:
                    atom["serial"] = atom_index
                    output["pdbqt"] += self.write_pdbqt_line(atom)
            else:
                line = template['original_line'][i]
                if branch_offset > 0 and (line.startswith("BRANCH") or line.startswith("ENDBRANCH")):
                    keyword, i, j = line.split()
                    i = int(i) + branch_offset
                    j = int(j) + branch_offset
                    line = "%s %3d %3d" % (keyword, i, j)
                output["pdbqt"] += line + os_linesep # e.g. BRANCH keywords

        output["atom_index"] = atom_index
        return output, success, error_msg

    @staticmethod
    def write_pdbqt_line(atom):
        return _write_pdbqt_line(atom["serial"], atom["xyz"][0], atom["xyz"][1], atom["xyz"][2],
                                 atom["partial_charges"], atom["name"], atom["resname"],
                                 atom["resid"], atom["atom_type"], atom["chain"],
                                 atom["alt_id"], atom["in_code"], atom["occupancy"],
                                 atom["temp_factor"], atom["record_type"])


    def write_pdbqt_string(self, flexres=()):
        ok = True
        err = ""
        pdbqt = {"rigid": "",
                 "flex":  {},
                 "flex_indices": []}
        atom_index = 0
        for res_id in set(flexres):
            output, ok_, err_ = self.write_flexres_from_template(res_id, atom_index)
            atom_index = output["atom_index"] # next residue starts here
            ok &= ok_
            err += err_
            pdbqt["flex_indices"].extend(output["flex_indices"])
            pdbqt["flex"][res_id] = ""
            pdbqt["flex"][res_id] += "BEGIN_RES %3s %1s%4d" % (res_id) + os_linesep
            pdbqt["flex"][res_id] += output["pdbqt"]
            pdbqt["flex"][res_id] += "END_RES %3s %1s%4d" % (res_id) + os_linesep

        # use non-flex lines for rigid part
        for i, atom in enumerate(self._atoms):
            if i not in pdbqt["flex_indices"] and atom["atom_type"] not in self.skip_types:
                pdbqt["rigid"] += self.write_pdbqt_line(atom)

        return pdbqt, ok, err

    @staticmethod
    def make_flexres_reactive(pdbqtstr, reactive_name, resname, prefix_atype="", residue_params=residue_params):
        atom_names = residue_params[resname]["atom_names"]
        bonds = residue_params[resname]["bonds"]
        def get_neigh(idx, bonds):
            neigh = set()
            for (i, j) in bonds:
                if i == idx:
                    neigh.add(j)
                elif j == idx:
                    neigh.add(i)
            return neigh
        react_idx = atom_names.index(reactive_name)
        one_bond_away = get_neigh(react_idx, bonds)
        two_bond_away = set()
        for i in one_bond_away:
            for j in get_neigh(i, bonds):
                if (j != react_idx) and (j not in one_bond_away):
                    two_bond_away.add(j)
        names_1bond = [atom_names[i] for i in one_bond_away]
        names_2bond = [atom_names[i] for i in two_bond_away]
        new_pdbqt_str = ""
        for i, line in enumerate(pdbqtstr.split(os_linesep)[:-1]):
            if line.startswith("ATOM") or line.startswith("HETATM"):
                name = line[12:16].strip()
                atype = line[77:].strip()
                if name == reactive_name:
                    new_type = prefix_atype + get_reactive_atype(atype, 1)
                elif name in names_1bond:
                    new_type = prefix_atype + get_reactive_atype(atype, 2)
                elif name in names_2bond:
                    new_type = prefix_atype + get_reactive_atype(atype, 3)
                else:
                    new_type = atype
                new_pdbqt_str += line[:77] + new_type + os_linesep
            else:
                new_pdbqt_str += line + os_linesep
        return new_pdbqt_str


    def atoms(self, atom_idx=None):
        """Return the atom i

        Args:
            atom_idx (int, list): index of one or multiple atoms

        Returns:
            ndarray: 2d ndarray (atom_id, atom_name, resname, resid, chainid, xyz, q, t)

        """
        if atom_idx is not None and self._atoms.size > 1:
            if not isinstance(atom_idx, (list, tuple, np.ndarray)):
                atom_idx = np.array(atom_idx, dtype=int)
            atoms = self._atoms[atom_idx]
        else:
            atoms = self._atoms

        return atoms.copy()

    def positions(self, atom_idx=None):
        """Return coordinates (xyz) of all atoms or a certain atom

        Args:
            atom_idx (int, list): index of one or multiple atoms (0-based)

        Returns:
            ndarray: 2d ndarray of coordinates (xyz)

        """
        return np.atleast_2d(self.atoms(atom_idx)['xyz'])

    def closest_atoms_from_positions(self, xyz, radius, atom_properties=None, ignore=None):
        """Retrieve indices of the closest atoms around a positions/coordinates
        at a certain radius.

        Args:
            xyz (np.ndarray): array of 3D coordinates
            raidus (float): radius
            atom_properties (str): property of the atoms to retrieve
                (properties: ligand, flexible_residue, vdw, hb_don, hb_acc, metal, water, reactive, glue)
            ignore (int or list): ignore atom for the search using atom id (0-based)

        Returns:
            ndarray: 2d ndarray (atom_id, atom_name, resname, resid, chainid, xyz, q, t)

        """
        index = self._KDTree.query_ball_point(xyz, radius, p=2, return_sorted=True)

        # When nothing was found around...
        if not index:
            return np.array([])

        # Handle the case when positions for of only one atom was passed in the input
        try:
            index = {i for j in index for i in j}
        except:
            index = set(index)

        if atom_properties is not None:
            if not isinstance(atom_properties, (list, tuple)):
                atom_properties = [atom_properties]

            try:
                for atom_property in atom_properties:
                    index.intersection_update(self._atom_annotations[atom_property])
            except:
                error_msg = 'Atom property %s is not valid. Valid atom properties are: %s'
                raise KeyError(error_msg % (atom_property, self._atom_annotations.keys()))

        if ignore is not None:
            if not isinstance(ignore, (list, tuple, np.ndarray)):
                ignore = [ignore]
            index = index.difference([i for i in ignore])

        index = list(index)
        atoms = self._atoms[index].copy()

        return atoms

    def closest_atoms(self, atom_idx, radius, atom_properties=None):
        """Retrieve indices of the closest atoms around a positions/coordinates
        at a certain radius.

        Args:
            atom_idx (int, list): index of one or multiple atoms (0-based)
            raidus (float): radius
            atom_properties (str or list): property of the atoms to retrieve
                (properties: ligand, flexible_residue, vdw, hb_don, hb_acc, metal, water, reactive, glue)

        Returns:
            ndarray: 2d ndarray (atom_id, atom_name, resname, resid, chainid, xyz, q, t)

        """
        return self.closest_atoms_from_positions(self._atoms[atom_idx]['xyz'], radius, atom_properties, atom_idx)

    def neighbor_atoms(self, atom_idx):
        """Return neighbor (bonded) atoms

        Args:
            atom_idx (int, list): index of one or multiple atoms (0-based)

        Returns:
            list_of_list: list of lists containing the neighbor (bonded) atoms (0-based)

        """
        if not isinstance(atom_idx, (list, tuple, np.ndarray)):
            atom_idx = [atom_idx]

        return [self._bonds[i] for i in atom_idx]



from rdkit import Chem
from .utils import mini_periodic_table


"""
create new RDKIT residue

mi  =  Chem.AtomPDBResidueInfo()
mi.SetResidueName('MOL')
mi.SetResidueNumber(1)
mi.SetOccupancy(0.0)
mi.SetTempFactor(0.0)

source: https://sourceforge.net/p/rdkit/mailman/message/36404394/
"""

from collections import namedtuple
PDBAtomInfo = namedtuple('PDBAtomInfo', "name resName resNum chain")

def getPdbInfoNoNull(atom):
    """extract information for populating an ATOM/HETATM line
    in the PDB"""
    # res = atom.GetResidue()
    minfo = atom.GetMonomerInfo()
    if minfo is None:
        atomic_number = atom.GetAtomicNum()
        if atomic_number == 0:
            name = '%-2s' % '*'
        else:
            name = '%-2s' % mini_periodic_table[atomic_number]
        chain = ' '
        resNum = 1
        resName = 'UNL'
    else:
        name = minfo.GetName()
        chain = minfo.GetChainId()
        resNum = minfo.GetResidueNumber()
        resName = minfo.GetResidueName()
    return PDBAtomInfo(name=name, resName=resName, resNum=resNum, chain=chain)


class Mol2MolSupplier():
    """ RDKit Mol2 molecule supplier.
    Parameters
        sanitize: perform RDKit sanitization of Mol2 molecule"""
    def __init__(self, filename, sanitize=True, removeHs=False, cleanupSubstructures=True):
        self.fp = open(filename, 'r')
        self._opts = {'sanitize':sanitize,
                'removeHs':removeHs,
                'cleanupSubstructures':cleanupSubstructures }
        self.buff = []

    def __iter__(self):
        return self

    def __next__(self):
        """ iterator step """
        while True:
            line = self.fp.readline()
            # empty line
            if not line:
                if len(self.buff):
                    # buffer full, returning last molecule
                    mol=Chem.MolFromMol2Block("".join(self.buff), **self._opts)
                    self.buff = []
                    return mol
                # buffer empty, stopping the iteration
                self.fp.close()
                raise StopIteration
            if '@<TRIPOS>MOLECULE' in line:
                # first molecule parsed
                if len(self.buff)==0:
                    self.buff.append(line)
                else:
                    # found the next molecule, breaking to return the complete one
                    break
            else:
                # adding another line in the current molecule
                self.buff.append(line)
        # found a complete molecule, returning it
        mol=Chem.MolFromMol2Block("".join(self.buff), **self._opts)
        self.buff = [line]
        return mol



############################################################################
#
# Authors: Stefano Forli, Ruth Huey
#
# Copyright: M. Sanner TSRI 2010
#
#############################################################################

#
# $Header: /opt/cvs/python/packages/share1.5/MolKit/WaterBuilder.py,v 1.1 2010/10/07 17:28:48 rhuey Exp $
#
# $Id: WaterBuilder.py,v 1.1 2010/10/07 17:28:48 rhuey Exp $
#

"""
This module implements the WaterBuilder class which adds water atom records to pdb files 

"""

import numpy as np

from .utils import geomutils


class WaterBuilder:
    """Class for adding waters to a pdb file.
    """

    def __init__(self, space=3.0, bond_dist=1.85, pcharge=0.000, residue="WAT", ATYPE="W", verbose=False):
        self.space = space
        self.bond_dist = bond_dist
        self.pcharge = pcharge
        self.residue = residue
        self.ATYPE = ATYPE
        self.verbose = verbose

    def atom_coord(self, atom):
        return list(map(float, atom[28:56].split()))

    def dist(self, firstline, secondline, precision=4):  
        coord1, coord2 = list(map(self.atom_coord, [firstline, secondline]))
        return np.round(np.sqrt((coord1[0]-coord2[0])**2+(coord1[1]-coord2[1])**2+(coord1[2]-coord2[2])**2), precision)

    def mean_pdb(self, firstline, secondline):
        # INFO   : calculate the mean point between two PDB atom lines
        # INPUT  : two pdb lines, an number for residue and atom numbering
        # OUTPUT : a pdb line
        coord1 = self.atom_coord(firstline)
        coord2 = self.atom_coord(secondline)
        x = (coord1[0]+coord2[0])/2
        y = (coord1[1]+coord2[1])/2
        z = (coord1[2]+coord2[2])/2
        atype = firstline[12:16]
        residue = "MEA"
        self.chain = "Y"
        count = 1
        index = 1
        #?SHOULD THIS BE atype from line 58?
        mean_atom = "ATOM  %5d %4s %3s %1s%4d    %8.3f%8.3f%8.3f  1.00 10.00          %1s" % (count, atype, residue, self.chain, index, x, y, z, self.ATYPE)
        #mean_atom = "ATOM  %5d %4s %3s %1s%4d    %8.3f%8.3f%8.3f  1.00 10.00          %1s" % (count, self.ATYPE, residue, self.chain, index, x, y, z, self.ATYPE)
        return mean_atom

    def closest(self, first_atom, atom_list, cutoff=99999):
        # INFO   : find the self.closest atom 
        # INPUT  : a PDB atom, a list of PDB atoms [, cutoff self.distance]
        # OUTPUT : the self.closest atom [=null, if cutoff not satisfied] and short self.distance found
        # EXTRA  : self.dist function required
        best_self.distance = 999999
        best_candidate = None
        for second_atom in atom_list:
            self.distance = self.dist(first_atom, second_atom)
            if self.distance < best_self.distance:
                best_self.distance = self.distance
                if best_self.distance < cutoff:
                    best_candidate = second_atom
        #if best_candidate != "":
        #    return best_candidate, best_self.distance
        #else:
        #    return best_candidate, best_self.distance
        return best_candidate, best_self.distance

    def mean3(self, firstline, secondline, thirdline):
        # INFO   : calculate the mean point between two PDB atom lines
        # INPUT  : two pdb lines, an number for residue and atom numbering
        # OUTPUT : a pdb line
        coord1 = self.atom_coord(firstline)
        coord2 = self.atom_coord(secondline)
        coord3 = self.atom_coord(thirdline)
        x = (coord1[0]+coord2[0]+coord3[0])/3
        y = (coord1[1]+coord2[1]+coord3[1])/3
        z = (coord1[2]+coord2[2]+coord3[2])/3
        atype = firstline[12:16]
        residue = "MEA"
        chain = "Y"
        count = 1
        index = 1
        mean_atom = "ATOM  %5d %4s %3s %1s%4d    %8.3f%8.3f%8.3f  1.00 10.00          %1s" % (count, self.ATYPE, residue, chain, index, x, y, z, self.ATYPE)
        return mean_atom

    def hydro(self, atom1, atom2, atom3=None, spacing=3):
        # INFO   : place a water atom W at given self.distance from an atom
        # INPUT  : (a) two pdb lines for N-H; (b) three pdb lines for -[N|O]- acceptor, -O-H donors; the spacing self.distance between the water and the atom
        # OUTPUT : one pdb line 
        #
        # Synopsis:
        #
        # -C-NH-C  => atom1 = N, atom2 = H
        # -C-N=C   => atom1 = C, atom2 = N, atom3 = C
        # -C-O-H   => atom1 = C, atom2 = O, atom3 = H

        #if not atom3:
        #    atype = line.split()[-1]
        if atom2.split()[-1]=="HD":
            spacing -= 1 # to put the W at 3A from the N-H
        index = 99 
        coord2 = self.atom_coord(atom2)
        x =(coord2[0])
        y =(coord2[1])
        z =(coord2[2])
        if self.verbose: print(x, y, z)
        #atype = "W"
        #residue="WAT"
        # residue 17:20
        # chain 21
        chain = atom1[21]
        residue = atom1[17:20]
        #chain="X"
        if atom3:
            atom4 = self.mean_pdb(atom1, atom3)
            vec_module = self.dist(atom2, atom4)
            coord1 = self.atom_coord(atom4) 
        else:
            coord1 = self.atom_coord(atom1)
            vec_module = self.dist(atom1, atom2)
        alpha = np.arccos((coord2[0]-coord1[0])/vec_module)    # x-axis angle
        beta  = np.arccos((coord2[1]-coord1[1])/vec_module)    # y-axis angle
        gamma = np.arccos((coord2[2]-coord1[2])/vec_module)    # z-axis angle
        wat_x = spacing*np.cos(alpha)+x
        wat_y = spacing*np.cos(beta)+y
        wat_z = spacing*np.cos(gamma)+z
        wet = "%s%5d %2s   %3s %1s%4d    %8.3f%8.3f%8.3f  1.00 10.00     %1.3f %1s\n" % (self.keyw, index, self.ATYPE, residue, chain, index, wat_x, wat_y, wat_z, self.pcharge, self.ATYPE)
        #wet = "%s%5d %2s   %3s %1s%4d    %8.3f%8.3f%8.3f  1.00 10.00     %1.3f %1s\n" % (self.keyw, index, self.ATYPE, self.residue, chain, index, wat_x, wat_y, wat_z, self.pcharge, self.ATYPE)
        return wet

    def hydroH(self, atom1, atom2, atom3):
        middle = self.hydro(atom1, atom2, spacing = 0)
        avg = self.mean_pdb(middle, atom3)
        last = self.hydro(avg, atom2, spacing=self.space)
        if self.verbose:
            print("middle=", middle)
            print("avg=", avg)
            print("last=", last)

    def vector(self, p1, p2=None):
        # accept both atoms and self.vectors as input
        #
        #if p1 == 0:
        #    coord1 = [ 0.0, 0.0, 0.0 ]
        #else:
        if self.verbose: print("received ", p1, p2)
        if type(p1) == type(str()):
            p1 = self.atom_coord(p1)
        x1,y1,z1 = p1
        if type(p2) == type(str()):
            p2 = self.atom_coord(p2)
        if not p2 == None:
            x2,y2,z2 = p2
            vec_x = x2-x1
            vec_y = y2-y1
            vec_z = z2-z1
            # it must be an array
            vec = np.array([vec_x, vec_y, vec_z], 'f')
            if self.verbose: print("REAL VECTOR", vec)
        else:
            vec = np.array([p1[0], p1[1], p1[2] ], 'f' )
            if self.verbose: print("ATOM VECTOR", vec)
        return vec

    def bound(self, atom, structure, exclude=None):
        """
        Identify all the atoms in "structure" that are @bond_dist from "atom"
        """
        bound_list = []
        tolerance = 0
        bond_dist = self.bond_dist
        if atom.split()[-1] == "HD":
            bond_dist = 1.15 # previous bond self.dist of 1.1 wouldn't be big enough for -S-H
            if self.verbose: print("HD mode")
        for candidate in structure:
            if candidate==atom or candidate==exclude:
                pass
            else:
                if candidate[0:4]=="ATOM" or candidate[0:6]=="HETATM":
                    if candidate.split()[-1]=="SA" or candidate.split()[-1]=="S":
                        tolerance = .35
                    else:
                        tolerance = 0
                    if self.dist(atom, candidate)<=bond_dist+tolerance:
                        if not candidate in bound_list:
                            bound_list.append(candidate)
                    else:
                        pass
        if len(bound_list) > 0:
            return bound_list
        else:
            print("ERROR: this atom seems to be disconnected:")
            print(atom)
            print("exit 3")
            exit(1)

    def furanbolic(self, atom, structure, max=2.35):
        # it should walk with pre-filtered co-planar atoms
        # HD's are automatically excluded
        the_ring = [atom]
        #print "I WAS REQUESTED TO OPERATE ON ", atom
        #print "I WILL SEARCH ON %d ATOMS"% len(structure)
        for item in structure:
            if item!=atom:
                if item.split()[-1]!="HD":
                    if self.dist(atom, item)<max:
                        the_ring.append(item)
                        if self.verbose:
                            print("APPEND")
                    #else:
                        #print "failed self.distance"
                        #print item
        #print "\n ########## FURANBOLIC CHECKPOINT ### \n"
        #print the_ring
        #print len(the_ring)
        if len(the_ring)==5:
            print(" - possible furan/oxazole found...")
            return True
        if len(the_ring)>6:
            print("WARNING: multiple atoms match the furan/oxazole check...")
            return True
        else:
            return False

    def Osp2(self, oxygen, atom1, atom2):
        waters = []
        # self.hydroxyl/ether mode
        angles=[120, -120]
        #angles = range(0, 360, 10)
        oxyvector = self.vector(oxygen, atom1)
        oxyvector = geomutils.normalize(oxyvector)
        residue = atom1[17:20]
        chain = atom1[21]
        for a in angles:
            roto = [oxyvector[0], oxyvector[1], oxyvector[2],  np.radians(a)]
            lone_pair_vector = self.vector(atom2, oxygen)
            lone_pair_vector = geomutils.normalize(lone_pair_vector)
            water =  geomutils.rotate_around_axis(-lone_pair_vector*self.space, roto, self.atom_coord(oxygen))  
            #residue = "99"
            #chain = "1"
            wet = "%s%5d %2s   %3s %1s%4d    %8.3f%8.3f%8.3f  1.00 10.00     %1.3f %1s\n"%(self.keyw, 1, self.ATYPE, residue, chain, 1, water[0], water[1], water[2], self.pcharge, self.ATYPE)
            #wet = "%s%5d %2s   %3s %1s%4d    %8.3f%8.3f%8.3f  1.00 10.00     %1.3f %1s\n"%(self.keyw, 1, self.ATYPE, residue, chain, 1, water[0], water[1], water[2], self.pcharge, self.ATYPE)
            waters.append(wet)
        if self.verbose: print("osp2 returning %d waters"%(len(waters))) 
        return waters

    def Osp2_NEW(self, oxygen, atom1, atom2):
        waters = []
        # self.hydroxyl/ether mode
        #angles = [120, -120]
        angles = list(range(0, 360, 10))
        oxyvector = self.vector(oxygen, atom1)
        oxyvector = self.vector(atom1, atom2)
        oxyvector = geomutils.normalize(oxyvector)
        mid = self.mean_pdb(atom1, atom2)
        residue = atom1[17:20]
        chain = atom1[21]
        for a in angles:
            roto = [oxyvector[0], oxyvector[1], oxyvector[2],  np.radians(a)]
            lone_pair_vector = self.vector(mid, oxygen)
            lone_pair_vector = geomutils.normalize(lone_pair_vector)
            water =  geomutils.rotate_around_axis(+lone_pair_vector*self.space, roto, self.atom_coord(oxygen))  
            #residue = "99"
            #chain = "1"
            # change to self.residue? replacing 99 by WAT??
            wet = "%s%5d %2s   %3s %1s%4d    %8.3f%8.3f%8.3f  1.00 10.00     %1.3f %1s\n"%(self.keyw, 1, self.ATYPE, residue, chain, 1, water[0], water[1], water[2], self.pcharge, self.ATYPE)
            waters.append(wet)
        if self.verbose: print("Osp2_NEW returning %d waters"%(len(waters))) 
        return waters

    def hydrate(self, mol, extended_atoms=False):
        added_waters = []
        atoms_list = []
        hydrate_list = []
        numbering_stuff = []
        water_mates = []
        self.EXTENDED_ATOMS = extended_atoms

        input = open(pdbqt, 'r').readlines()
        for line in input:
            if line[0:4] == "ATOM" or line[0:6] == "HETATM":
                atype = line.split()[-1]
                atoms_list.append(line)
                if atype == "OA" or atype == "NA" or atype == "HD":
                    hydrate_list.append(line)
            if line[0:4] == "ATOM":
                self.keyw = "ATOM  "
            if line[0:6] == "HETATM":
                self.keyw = "HETATM"

        # Scan the list to add waters
        for atom in hydrate_list:
            atype = atom.split()[-1]
            if self.verbose:
                print("PROCESSING ATOM :", atom)
                print("atype = ", atype)
            HYDROXYL = False
            # ordinal position in the original file
            position = int(atom.split()[1])
            # add the atom to be hydrated to the buffer list
            waters_generated = [atom]  #@@reinitialize waters_generated here
            if self.verbose:
                print("processing ", atom)
                print("0: water_mates=", water_mates)
                print("1: len(waters_generated)=", len(waters_generated))
                print("2:  len(water_mates=", len(water_mates))
            # find the master atom(s)
            master = self.bound(atom, atoms_list)
            if len(master) == 0:
                print("\n\nERROR: this atom is disconnected:\n", atom)
                #??return() instead??@@
                print("exit 0")
                exit(1)
            # HYDROGENS #####################################################################################
            #print "before HD , len(master)=", len(master), " atype=", atype
            if atype == "HD":
                # check for errors
                if len(master) > 1:
                    print("\n\nERROR (HD) : there is a proximity error and the following hydrogen is in close contact with more than one atom")
                    print(atom)
                    print("Bound mates:")
                    for m in master:
                        print(m[:-1]," ==>", self.dist(m,atom))
                    #??return() instead??@@
                    print("exit 1")
                    exit(1)
                else:
                    # calculate the Water vector
                    wet = self.hydro(master[0], atom) #@@REPAIR THIS apparent duplicate!!!
                    if self.verbose: 
                        print("after call to hydro, wet=",wet)
                    waters_generated.append(wet)
                    if self.verbose:
                        print("HD: added wet: now %d waters_generated:" %len(waters_generated))
                        for ind in waters_generated: print(ind)
                        print("HD: added w_g to w_m: previously %d water_mates:" %len(water_mates))
                        for ind in water_mates: print(ind)
                        print("HD: 2: %d waters_generated:" %len(waters_generated))
                        for ind in waters_generated: print(ind)
                water_mates.append(waters_generated)
                if self.verbose: print("HD: 2: NOW %d water_mates:" %len(water_mates))
                #numbering_stuff has an entry for for each entry in water_mates
                numbering_stuff.append([position, (len(waters_generated)-1)] )

            # OXYGENS #####################################################################################
            if atype == "OA":
                # OA includes the following options:
                #
                ## Two mates
                #  ---------
                #
                # -X-OA-X-    (ethers)
                #
                # -X-OA-HD    (hydroxyls)
                #
                # [-X-OA-X-]  (furan/pyran like)
                #
                #
                ## One mate
                # ----------
                #
                # -X=O        (carbonyl, carboxylate, nitro)   [ DONE]
                #
                #
                #  |
                # -X=O        (sulphonyl, phosphonyl) 
                #  |
                #
                if len(master)==1:
                    # # identify the mates of the master
                    MASTER = master[0]
                    residue = MASTER[17:20]
                    chain = MASTER[21]
                    mates = self.bound(MASTER, atoms_list, exclude=atom)
                    #print "masters", master
                    #print "mates", mates
                    # Phosphates check
                    if len(mates) <=2:
                        # 1. calculate the plane
                        v12 = self.vector(atom, MASTER)
                        v23 = self.vector(mates[0],MASTER) 
                        #plane = np.cross(v12, v23)
                        plane0,plane1,plane2 = geomutils.normalize(np.cross(v12,v23))
                        self.chain = 1 # TODO read this from the atom
                        #keyw, index, ATYPE, residue, chain, pcharge = "ATOM  ", 1, "C", "RES", 1, 0.0000
                        #print "CARBONYL"
                        #print "%s%5d %2s   %3s %1s%4d    %8.3f%8.3f%8.3f  1.00 10.00     %1.3f %1s" % (keyw, index, ATYPE, residue, chain, index, plane[0], plane[1], plane[2], pcharge, ATYPE)

                        # O-lone pair 1
                        roto = [ plane0, plane1, plane2, np.radians(50) ]
                        wat = geomutils.rotate_around_axis(geomutils.normalize(-v12)*self.space, roto, self.atom_coord(atom))
                        wet = "%s%5d %2s   %3s %1s%4d    %8.3f%8.3f%8.3f  1.00 10.00     %1.3f %1s\n" % (self.keyw, 1, self.ATYPE, residue, chain, 1, wat[0], wat[1], wat[2], self.pcharge, self.ATYPE)
                        #wet = "%s%5d %2s   %3s %1s%4d    %8.3f%8.3f%8.3f  1.00 10.00     %1.3f %1s\n" % (self.keyw, 1, self.ATYPE, self.residue, self.chain, 1, wat[0], wat[1], wat[2], self.pcharge, self.ATYPE)
                        if self.verbose:
                            print("O1: wet:") 
                            print(wet)
                        waters_generated.append(wet)
                        if self.verbose:
                            print("waters_generated:")
                            print(waters_generated)
                            print("O1: waters_generated=", end=' ') 
                            for ind in waters_generated: print(ind)
            
                        # O-lone pair 2
                        roto = [ plane0, plane1, plane2, np.radians(-50) ]
                        if self.verbose: print("calling rotate_around_axis with self.atom_coord(", atom, "), roto=",roto)
                        wat = geomutils.rotate_around_axis(geomutils.normalize(-v12)*self.space, roto, self.atom_coord(atom))
                        if self.verbose: print("wat =", wat)
                        wet = "%s%5d %2s   %3s %1s%4d    %8.3f%8.3f%8.3f  1.00 10.00     %1.3f %1s\n" % (self.keyw, 1, self.ATYPE, residue, chain, 1, wat[0], wat[1], wat[2], self.pcharge, self.ATYPE)
                        #wet = "%s%5d %2s   %3s %1s%4d    %8.3f%8.3f%8.3f  1.00 10.00     %1.3f %1s\n" % (self.keyw, 1, self.ATYPE, self.residue, self.chain, 1, wat[0], wat[1], wat[2], self.pcharge, self.ATYPE)
                        waters_generated.append(wet)
                        if self.verbose:
                            print("O2: %d wet: %s" %(len(wet), wet))
                            print("O2: %d waters_generated="%(len(waters_generated)), end=' ') 
                            for ind in waters_generated: print(ind)
                        water_mates.append(waters_generated)
                        if self.verbose:
                            print("O2: %d water_mates="%(len(water_mates)), end=' ') 
                            for ind in water_mates: print(ind)
                        numbering_stuff.append([int(position), len(waters_generated)-1])
                        #print waters_generated
                    else:
                        if self.EXTENDED_ATOMS:
                            #if self.verbose:
                                #print "master", master[0]
                                #print "mates", mates
                                #print atom
                                #print "Possible phosphate/sulphate atom... I can't stand Phosphates... GIVING UP"
                                #print "\n\n[ we're not equipped to manage those atoms... yet ]"
                                #print "\n\n[ we're testing to manage those atoms...EXPERIMENTAL!  ]"
                            #chain = 1
                            #self.residue = 1 
                            #print "\n\n  ############## TESTING FACILITY #########"
                            directive = self.vector(master[0],atom)
                            directive = geomutils.normalize(directive)*self.space
                            #if self.verbose: print "%s%5d %2s   %3s %1s%4d    %8.3f%8.3f%8.3f  1.00 10.00     %1.3f %1s\n" % (keyw, 1, self.ATYPE, residue, chain, 1, directive[0], directive[1], directive[2], pcharge, self.ATYPE)

                            for q in mates:
                                #print q
                                position_q = self.vector(q) 
                                position_q = geomutils.vecSum(position_q, directive)
                                push = geomutils.normalize(self.vector(atom, position_q))
                                start = self.vector(atom)
                                lpair = geomutils.vecSum(start, push*self.space)
                                wet = "%s%5d %2s   %3s %1s%4d    %8.3f%8.3f%8.3f  1.00 10.00     %1.3f %1s\n" % (self.keyw, 1, self.ATYPE, residue, chain, 1, lpair[0], lpair[1], lpair[2], self.pcharge, self.ATYPE)
                                #wet = "%s%5d %2s   %3s %1s%4d    %8.3f%8.3f%8.3f  1.00 10.00     %1.3f %1s\n" % (self.keyw, 1, self.ATYPE, self.residue, self.chain, 1, lpair[0], lpair[1], lpair[2], self.pcharge, self.ATYPE)
                                if self.verbose:
                                    print("EA: wet=", end=' ') 
                                    for ind in wet: print(ind)
                                #position_q = normalize(position_q)
                                #ACTUAL = normalize(-vector(atom, position_q))*space
                                waters_generated.append(wet)
                                if self.verbose:
                                    print("EA: waters_generated=", end=' ') 
                                    for ind in waters_generated: print(ind)
                                #if self.verbose: print "EXT ATOMS: waters_generated=", waters_generated
                            water_mates.append(waters_generated)
                            if self.verbose: 
                                print("661:EA: %d water_mates:"%(len(water_mates)), end=' ') 
                                for ind in water_mates: 
                                    for j in ind:
                                        print(j)
                            #if self.verbose: print "         : water_mates=", water_mates
                            numbering_stuff.append([int(position), len(waters_generated)-1])
                            #print "\n\n  ############## TESTING FACILITY #########"
                            #exit(1)
                if len(master) == 2:
                    #print "---[ MODE 2 ]---"
                    for m in master:
                        if m.split()[-1] == "HD":
                            HYDROXYL = True
                            #print " >>> Found hydroxyl"
                    if not HYDROXYL:
                        # calculate the plane formed by oxygen and the two masters
                        O_plane = geomutils.calcPlane(atom, master[0], master[1])
                        # get the list of all the co-planar atoms in the structure
                        coplanar_mates = geomutils.coplanar(O_plane, atoms_list, atom)
                        #print "COPLANAR_MATES", len(coplanar_mates)
                        #print "FURANBOLIC", furanbolic(atom, coplanar_mates)
                        # check if there are at least 4 coplanar atoms to make a ring with the OA
                        if len(coplanar_mates) >=4 and self.furanbolic(atom, coplanar_mates):    
                            wet = self.hydro(master[0], atom, master[1]) #2: @@REPAIR THIS apparent duplicate!!!
                            if self.verbose:
                                print("notHYDROXYL: wet=", end=' ') 
                                for ind in wet: print(ind)
                            waters_generated.append(wet)
                            if self.verbose:
                                print("notHYDROXYL: waters_generated=", end=' ') 
                                for ind in waters_generated: print(ind)
                            #if self.verbose:
                            #    print "NOT_HYDROXYL: waters_generated=", waters_generated
                            #print "FURAN MODE"
                        else:
                            if self.verbose: print("FURAN MODE FAILED")
                            lp_waters = self.Osp2(atom, master[0], master[1])
                            if self.verbose:
                                print("lp_waters:", end=' ')
                                for ind in lp_waters: print(ind)
                            for w in lp_waters:
                                waters_generated.append(w)
                            if self.verbose:
                                print("else in not HYDROXYL: waters_generated=", end=' ') 
                                for ind in waters_generated: print(ind)
                                print("NOT_COPLANAR ELSE 1: waters_generated=", waters_generated)
                    else:
                        #print "HYDROXYL/ETHER MODE"
                        lp_waters = self.Osp2(atom, master[0], master[1])
                        if self.verbose: print("lp_waters:")
                        for w in lp_waters:
                            if self.verbose:print(w)
                            waters_generated.append(w)
                        if self.verbose:
                            print("ELSE: waters_generated=", end=' ') 
                            for ind in waters_generated: print(ind)
                            print("before adding else of not HYDROXYL, water_mates=", end=' ')
                            for ind in water_mates: print(ind)
                    water_mates.append(waters_generated)
                    if self.verbose:
                        print("LAST 1: waters_generated=", end=' ') 
                        for ind in waters_generated: print(ind)
                        print("LAST 1: water_mates =", end=' ') 
                        for wm in water_mates : print(wm)
                        print("LAST 1: waters_generated=", waters_generated)
                        print("LAST 1: water_mates =", water_mates) 
                    numbering_stuff.append([int(position), len(waters_generated)-1])

            # NITROGEN #####################################################################################
            if atype == "NA":
                if len(master) == 1:
                    # calculate the Water vector
                    wet = self.hydro(master[0], atom) #3: @@REPAIR THIS APPARENT DUPLICATE??
                    waters_generated.append(wet)
                    water_mates.append(waters_generated)
                    if self.verbose:
                        print("NA 1: wet=", end=' ') 
                        for ind in wet: print(ind)
                        print("NA 1: waters_generated=", end=' ') 
                        for ind in waters_generated: print(ind)
                        print("NA 1: water_mates=", end=' ') 
                        for ind in water_mates: print(ind)
                    numbering_stuff.append([int(position), len(waters_generated)-1])

                # nitrile mode
                if len(master) == 2:
                    wet = self.hydro(master[0], atom, master[1]) #4: @@REPAIR THIS APPARENT DUPLICATE4??
                    if self.verbose: print("nitrile mode: wet:", wet)
                    #for ind in wet: print ind
                    waters_generated.append(wet)
                    if self.verbose: 
                        print("N: waters_generated:", end=' ')
                        for ind in waters_generated: print(ind)
                    water_mates.append(waters_generated)
                    if self.verbose: 
                        print("N: water_mates:", end=' ')
                        for ind in water_mates: print(ind)
                        print("NA 2: wet=", wet)
                        print("NA 2: waters_generated=", waters_generated)
                        print("NA 2: water_mates=", water_mates)
                    numbering_stuff.append([int(position), len(waters_generated)-1])
                    if self.verbose:
                        print("nitrile numbering_stuff:", end=' ')
                        for ind in numbering_stuff: print(ind)

                # tertiary amine HB receptor
                if len(master) == 3: 
                    #print "EXPERIMENTAL! potential tertiary amine"
                    master_center = self.mean3(master[0], master[1], master[2])
                    wet = self.hydro(master_center, atom) #5: @@REPAIR THIS APPARENT DUPLICATE5??
                    if self.verbose: 
                        print("tertiary amine: wet=", end=' ')
                        for ind in wet: print(ind)
                    waters_generated.append(wet)
                    if self.verbose: 
                        print("tertiary amine: waters_generated=", end=' ')
                        for ind in waters_generated: print(ind)
                    water_mates.append(waters_generated)
                    if self.verbose: 
                        print("tertiary amine: water_mates=", end=' ')
                        for ind in water_mates: print(ind)
                        print("NA 3: wet=", wet)
                        print("NA 3: waters_generated=", waters_generated)
                        print("NA 3: water_mates=", water_mates)
                    numbering_stuff.append([int(position), len(waters_generated)-1])
                    if self.verbose: print(numbering_stuff[-1], ' added to numbering_stuff')


        ctr = 0
        for mates in water_mates:
            index = input.index(mates[0])
            if self.verbose: 
                print("mates=", mates)
                print("mates[0]=", mates[0])
            line = ""
            for atom in mates:
                line += atom
            if self.verbose: print("@@ %d input line was = %s"%(index,  input[index]))
            input[index] = line
            if self.verbose:
                print("now %d input line set to %s"%(index,  line))
                print("~~~numbering_stuff[%d]=%s"%(ctr, numbering_stuff[ctr]))
            ctr+=1
            #if index==20: raise 'abc'
        count = 1
        # nicely split and format the lines that need to be splitted from
        # the previous addition of water molecules
        final = []
        if self.verbose: print("len(input)=", len(input), ' len(final)=', len(final))
        for line in input:
            line = line.split("\n")
            if self.verbose:
                print("ZZZZ: line=", end=' ') 
                for  YYYY in line: print(YYYY)
            if len(line)==3 and line[-1]!='': 
                #raise 'THREE!'
                line = line.remove(line[1])
            ct = 0
            for item in line:
                #@@@FIX THIS HACK!!!
                if not item == "" and item.find("WAT")<0 and item.find("W    99 1")<0:
                    final.append(item)
                    if len(item)>13 and item[13]=='W':
                        if self.verbose: print("after adding W ", item," len(final)=", len(final))
                    ct += 1
        if self.verbose: print("len(input)=", len(input), ' len(final)=', len(final))

        # renumbering the PDBQT atoms
        for line in final:
            if self.verbose: print("at top line=", line, " contains WAT ", line.find("WAT")>0)
            if line.find("WAT")<0 and (line[0:4] == "ATOM" or line[0:6] == "HETATM"):
                value = "%4s" % count
                idx = final.index(line)
                final[idx] = line[0:7]+value+line[11:]
                count += 1
                if self.verbose: print("in loop line=", line)

        # process the BRANCH numbers
        for line in final:
            idx = final.index(line)
            if "BRANCH" in line:
                #print "Processing line => ", line
                line = line.split()
                value1, value2 = int(line[1]), int(line[2])
                #print "BEFORE =>", value1, value2 
                addendum1 = 0
                addendum2 = 0
                for mark in numbering_stuff:
                    #print "Mark is :\t POSITION:", mark[0], " COUNT", mark[1]
                    if value1 > mark[0]:
                        addendum1 += mark[1]
                        #print "value 1 is bigger than mark[0]:", value1, " | ", mark[0]
                    if value2 > mark[0]:
                        addendum2 += mark[1]
                        #print "value 2 is bigger than mark[0]:", value2, " | ", mark[0]
                value1 += addendum1
                value2 += addendum2
                #print "AFTER  =>", value1, value2
                #print "- - - - - "
                final[idx] = line[0]+" "+str(value1)+" "+str(value2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko bond typer
#

import os
import sys
from collections import defaultdict
from operator import itemgetter


class BondTyperLegacy:

    def __call__(self, setup, flexible_amides, rigidify_bonds_smarts, rigidify_bonds_indices, not_terminal_atoms=[]):
        """Typing atom bonds in the legacy way

        Args:
            setup: MoleculeSetup object

            rigidify_bond_smarts (list): patterns to freeze bonds, e.g. conjugated carbons
        """
        def _is_terminal(idx):
            """ check if the atom has more than one connection with non-ignored atoms"""
            if setup.get_element(idx) == 1:
                return True
            return len([x for x in setup.get_neigh(idx) if not setup.get_ignore(x)]) == 1
        amide_bonds = [(x[0], x[1]) for x in setup.find_pattern('[NX3]-[CX3]=[O,N]')] # includes amidines

        # tertiary amides with non-identical substituents will be allowed to rotate
        tertiary_amides = [x for x in setup.find_pattern('[NX3]([!#1])([!#1])-[CX3]=[O,N]')]
        equivalent_atoms = setup.get_equivalent_atoms()
        num_amides_removed = 0
        num_amides_originally = len(amide_bonds)
        for x in tertiary_amides:
            r1, r2 = x[1], x[2]
            if equivalent_atoms[r1] != equivalent_atoms[r2]:
                amide_bonds.remove((x[0], x[3]))
                num_amides_removed += 1
        assert(num_amides_originally == num_amides_removed + len(amide_bonds))

        to_rigidify = set()
        n_smarts = len(rigidify_bonds_smarts)
        assert(n_smarts == len(rigidify_bonds_indices))
        for i in range(n_smarts):
            a, b = rigidify_bonds_indices[i]
            smarts = rigidify_bonds_smarts[i]
            indices_list = setup.find_pattern(smarts)
            for indices in indices_list:
                atom_a = indices[a]
                atom_b = indices[b]
                to_rigidify.add((atom_a, atom_b))
                to_rigidify.add((atom_b, atom_a))

        for bond_id, bond_info in setup.bond.items():
            rotatable = True
            bond_order = setup.bond[bond_id]['bond_order']
            # bond requested to be rigid
            if bond_id in to_rigidify:
                bond_order = 1.1 # macrocycle class breaks bonds if bond_order == 1
                rotatable = False
            # non-rotatable bond
            if bond_info['bond_order'] > 1:
                rotatable = False
            # in-ring bond
            if len(bond_info['in_rings']):
                rotatable = False
            # it's a terminal atom (methyl, halogen, hydrogen...)
            is_terminal_1 = _is_terminal(bond_id[0]) and (bond_id[0] not in not_terminal_atoms)
            is_terminal_2 = _is_terminal(bond_id[1]) and (bond_id[1] not in not_terminal_atoms)
            if is_terminal_1 or is_terminal_2:
                rotatable = False
            # check if bond is amide
            # NOTE this should have been done during the setup, right?
            if (bond_id in amide_bonds or (bond_id[1], bond_id[0]) in amide_bonds) and not flexible_amides:
                rotatable = False
                bond_order = 99
            setup.bond[bond_id]['rotatable'] = rotatable
            setup.bond[bond_id]['bond_order'] = bond_order




from meeko import RDKitMolCreate
from meeko import PDBQTMolecule
from meeko import MoleculePreparation
from meeko import PDBQTWriterLegacy
from rdkit import Chem
import pathlib

workdir = pathlib.Path(__file__)
datadir = workdir.parents[0] / "rdkitmol_from_docking_data"

# test for reasonable bond lengths. This will fail if the atom names
# in RDKiMolCreate.flexres[<resname>]["atom_names"] are in the wrong
# order w.r.t RDKitMolCreate.flexres[<resname>]["smiles"]
bond_range_H = [0.85, 1.15]
bond_range_no_H = [1.16, 1.65]
bond_S_bonus = 0.2

def check_rdkit_bond_lengths(fpath, nr_expected_none, is_dlg, skip_typing):
    pdbqtmol = PDBQTMolecule.from_file(fpath, is_dlg=is_dlg, skip_typing=skip_typing)
    return run_from_pdbqtmol(pdbqtmol, nr_expected_none)

def run_from_pdbqtmol(pdbqtmol, nr_expected_none=0):
    mols = RDKitMolCreate.from_pdbqt_mol(pdbqtmol) 
    assert(mols.count(None) == nr_expected_none)
    nr_conformers = set()
    for mol in mols:
        if mol is None:
            continue
        nr_conformers.add(mol.GetNumConformers())
        assert(len(nr_conformers) == 1)
        for confid in range(list(nr_conformers)[0]):
            positions = mol.GetConformer(confid).GetPositions()
            for bond in mol.GetBonds():
                has_H = bond.GetBeginAtom().GetAtomicNum() == 1
                has_H = has_H or (bond.GetEndAtom().GetAtomicNum() == 1)
                bond_range = bond_range_H if has_H else bond_range_no_H
                has_S = bond.GetBeginAtom().GetAtomicNum() in [16, 17, 35]
                has_S = has_S or (bond.GetEndAtom().GetAtomicNum() in [16, 17, 35])
                if has_S:
                    bond_range = [value + bond_S_bonus for value in bond_range] 
                a = positions[bond.GetBeginAtomIdx(), :]
                b = positions[bond.GetEndAtomIdx(), :]
                dist = sum([(a[i]-b[i])**2 for i in range(3)])**0.5
                assert(dist > bond_range[0])
                assert(dist < bond_range[1])

def test_asn_phe():
    fpath = datadir / "macrocycle-water-asn-phe.pdbqt"
    check_rdkit_bond_lengths(fpath, nr_expected_none=0, is_dlg=False, skip_typing=True)

def test_22_flexres():
    fpath = datadir / "22-flexres.pdbqt"
    check_rdkit_bond_lengths(fpath, nr_expected_none=0, is_dlg=False, skip_typing=True)

def test_phe_badphe():
    fpath = datadir / "arg_gln_asn_phe_badphe.pdbqt"
    check_rdkit_bond_lengths(fpath, nr_expected_none=1, is_dlg=False, skip_typing=True)

def test_arg_his():
    fpath = datadir / "arg_his.pdbqt"
    check_rdkit_bond_lengths(fpath, nr_expected_none=0, is_dlg=False, skip_typing=True)


# The following tests  generate the PDBQT and convert it back to RDKit,
# as opposed to the tests above which start from PDBQT.

mk_prep = MoleculePreparation()
mk_prep_wet = MoleculePreparation(hydrate=True)

def run(sdfname, wet=False):
    fpath = datadir / sdfname
    for mol in Chem.SDMolSupplier(str(fpath), removeHs=False):
        if wet:
            setups = mk_prep_wet.prepare(mol)
            pdbqt, is_ok, error_msg = PDBQTWriterLegacy.write_string(setups[0])
        else:
            setups = mk_prep.prepare(mol)
            pdbqt, is_ok, error_msg = PDBQTWriterLegacy.write_string(setups[0])
        pmol = PDBQTMolecule(pdbqt)
        run_from_pdbqtmol(pmol)

def test_small_01_zero_deuterium(): run("small-01_zero-deuterium.sdf")
def test_small_01_one_deuterium(): run("small-01_one-deuterium.sdf")
def test_small_01_two_deuterium(): run("small-01_two-deuterium.sdf")
def test_small_01_three_deuterium(): run("small-01_three-deuterium.sdf")
def test_small_01_four_deuterium(): run("small-01_four-deuterium.sdf")

def test_small_02_zero_deuterium(): run("small-02_zero-deuterium.sdf")
def test_small_02_one_deuterium_A(): run("small-02_one-deuterium-A.sdf")
def test_small_02_one_deuterium_B(): run("small-02_one-deuterium-B.sdf")
def test_small_02_one_deuterium_C(): run("small-02_one-deuterium-C.sdf")
def test_small_02_two_deuterium_A(): run("small-02_two-deuterium-A.sdf")
def test_small_02_two_deuterium_B(): run("small-02_two-deuterium-B.sdf")
def test_small_02_two_deuterium_C(): run("small-02_two-deuterium-C.sdf", wet=True)
def test_small_02_two_deuterium_D(): run("small-02_two-deuterium-D.sdf", wet=True)
def test_small_02_five_deuterium(): run("small-02_five-deuterium.sdf", wet=True)

def test_small_03_zero_deuterium(): run("small-03_zero-deuterium.sdf", wet=True)
def test_small_03_one_deuterium_A(): run("small-03_one-deuterium-A.sdf", wet=True)
def test_small_03_one_deuterium_B(): run("small-03_one-deuterium-B.sdf", wet=True)
def test_small_03_two_deuterium(): run("small-03_two-deuterium.sdf")
def test_small_03_three_deuterium(): run("small-03_three-deuterium.sdf")

def test_small_04(): run("small-04.sdf", wet=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Raccoon
#

import math
import sys

import numpy as np


def get_vector(coor1, coord2):
    """ calculate normalized vector between atoms"""
    vec = np.array([coord2[0] - coord1[0], coord2[1] - coord1[1], coord2[2] - coord1[2]], 'f')
    return normalize(vec)


def vector(a, b):
    """
    Return the vector between a and b
    """
    return b - a

def resize_vector(v, length, origin=None):
    """ Resize a vector v to a new length in regard to a origin """
    if origin is not None:
        return (normalize(v - origin) * length) + origin
    else:
        return normalize(v) * length


def get_vector_normal(vector):
    """return the first vector normal to the input vector"""
    return np.array(vector[1], vector[0], vector[2])


def normalize(v):
    """ numpy normalize"""
    return v / np.sqrt(np.dot(v, v))

###def centroid(self, atomlist):
###    """ calculate centroid """
###    centroid =  np.array([0., 0., 0.], 'f')
###    for i in atomlist:
###        centroid += obutils.getAtomCoord(p.getAtomCoord(i)
###    return centroid/len(atomlist)
###

def averageCoords(coordList):
    """ http://stackoverflow.com/questions/23020659/fastest-way-to-calculate-the-centroid-of-a-set-of-coordinate-tuples-in-python-wi"""
    avg = np.zeros(3)
    for c in coordList:
        avg += c
    return avg / len(coordList)


def calcPlaneVect(v1, v2, norm=True):
    """ calculate plane defined by two numpy.vectors"""
    # print "PLANE", v1, v2
    plane = np.cross(v1, v2)
    if not norm:
        return plane
    return normalize(plane)


def rotation_matrix(axis, theta):
    """
    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by theta radians.
    http://stackoverflow.com/questions/6802577/python-rotation-of-3d-vector
    """
    axis = np.asarray(axis)
    theta = np.asarray(theta)
    axis = axis/math.sqrt(np.dot(axis, axis))
    a = math.cos(theta/2)
    b, c, d = -axis*math.sin(theta/2)
    aa, bb, cc, dd = a*a, b*b, c*c, d*d
    bc, ad, ac, ab, bd, cd = b*c, a*d, a*c, a*b, b*d, c*d
    return np.array([[aa+bb-cc-dd, 2*(bc+ad), 2*(bd-ac)],
                     [2*(bc-ad), aa+cc-bb-dd, 2*(cd+ab)],
                     [2*(bd+ac), 2*(cd-ab), aa+dd-bb-cc]])


def rotate_around_axis(vector, rot_axis, apply_point=[0., 0., 0.]):
    """
    Rotate a vector around an axis (rot_axis) applied to the point apply_point (NEW)

    Rotate a vector applied in apply_point around a pivot rot_axis ?
    vector = vector that is rotated
    rot_axis = vector around wich rotation is performed

        ?????? CHANGING THE INPUT VALUE?
    """
    # From Ludo
    # vector
    x = vector[0]
    y = vector[1]
    z = vector[2]

    # rotation pivot
    u = rot_axis[0]
    v = rot_axis[1]
    w = rot_axis[2]
    ux = u*x
    uy = u*y
    uz = u*z
    vx = v*x
    vy = v*y
    vz = v*z
    wx = w*x
    wy = w*y
    wz = w*z
    sa = math.sin(rot_axis[3])
    ca = math.cos(rot_axis[3])
    #vector[0]=(u*(ux+vy+wz)+(x*(v*v+w*w)-u*(vy+wz))*ca+(-wy+vz)*sa)+ apply_point[0]
    #vector[1]=(v*(ux+vy+wz)+(y*(u*u+w*w)-v*(ux+wz))*ca+(wx-uz)*sa)+ apply_point[1]
    #vector[2]=(w*(ux+vy+wz)+(z*(u*u+v*v)-w*(ux+vy))*ca+(-vx+uy)*sa)+ apply_point[2]
    p0 = (u*(ux+vy+wz)+(x*(v*v+w*w)-u*(vy+wz))*ca+(-wy+vz)*sa) + apply_point[0]
    p1 = (v*(ux+vy+wz)+(y*(u*u+w*w)-v*(ux+wz))*ca+(wx-uz)*sa) + apply_point[1]
    p2 = (w*(ux+vy+wz)+(z*(u*u+v*v)-w*(ux+vy))*ca+(-vx+uy)*sa) + apply_point[2]
    #b = [vector, m, rot_axis]

    return np.array([p0, p1, p2])


def rotation_axis(p0, p1, p2, origin=None):
    """
    Compute rotation axis centered at the origin if not None
    """
    r = normalize(np.cross(vector(p1, p0), vector(p2, p0)))

    if origin is not None:
        return origin + r

    return p0 + r


def atom_to_move(o, p):
    """
    Return the coordinates xyz of an atom just above acceptor/donor atom o
    """
    # It will not work if there is just one dimension
    p = np.atleast_2d(p)
    return o + normalize(-1. * vector(o, np.mean(p, axis=0)))


def rotate_point(p, p1, p2, angle):
    """ Rotate the point p around the axis p1-p2
    Source: http://paulbourke.net/geometry/rotate/PointRotate.py"""
    # Translate the point we want to rotate to the origin
    pn = p - p1

    # Get the unit vector from the axis p1-p2
    n = p2 - p1
    n = normalize(n)

    # Setup the rotation matrix
    c = np.cos(angle)
    t = 1. - np.cos(angle)
    s = np.sin(angle)
    x, y, z = n[0], n[1], n[2]

    R = np.array([[t*x**2 + c, t*x*y - s*z, t*x*z + s*y],
                 [t*x*y + s*z, t*y**2 + c, t*y*z - s*x],
                 [t*x*z - s*y, t*y*z + s*x, t*z**2 + c]])

    # ... and apply it
    ptr = np.dot(pn, R)

    # And to finish, we put it back
    p = ptr + p1

    return p


def getVecNormalToVec(vec):
    """
    calculate a vector that is normal to the numpy.array vector input

    Source: http://forums.create.msdn.com/forums/p/9551/50048.aspx
    A coworker pointed out a trick to get a vector perpendicular to the normal vector:
    simply swap two of the values, negate one of those, and zero the third.
    So, if I have a normal vector of form Vector3(a, b, c), then one such vector that
    is perpendicular to it is Vector3(b, -a, 0).  Thus, there are six possible vectors
    that are attainable by this method.  The only trouble case is when the normal vector
    contains elements whose values are zero, in which case you have to be a bit careful
    which values you swap and negate.  You just never want to end up with the zero vector.

    n n 0
    n 0 n
    0 n n
    n 0 0
    0 n 0
    n n n
    0 0 0

    if np.array([0., 0., 0.], 'f') == vec:
        print "Warning: zero vector, no normal possible"
        return vec
    x,y,z = vec
    zero = [ a == 0 for a in vec ]
    idx = range(3)
    swap1, swap2, zero = None, None, None
    for i in range(3):
        if not vec[i] == 0:
            negate = -vec[i]
        else:
            swap1 =
    """
    if (not vec[1] == 0) or (not vec[2] == 0):
        c = np.array([1.,0.,0.], 'f')
    else:
        c = np.array([0.,1.,0.], 'f')
    return calcPlane(vec, c, norm=True)


def calcDihedral(a1, a2, a3, a4):
    """ given 4 set of coordinates return the dihedral
        angle between them
    """
    v1 = vector(a1, a2)
    v2 = vector(a3, a2)
    v3 = vector(a3, a4)

    v4 = np.cross(v1, v2)
    v5 = np.cross(v2, v4)
    try:
        dihe = math.atan2(np.dot(v3,v4), np.dot(v3,v5) * math.sqrt(np.dot(v2,v2)))
    except ZeroDivisionError:
        dihe = 0.
    return dihe


def makeCircleOnPlane(center, r, normal, points = 8):
    """
    Calculate the points of a circle lying on an arbitrary plane
    defined by the normal vector.
    center : coords of center of the circle
    r      : radius
    normal : normal of the plane where the circle lies
    points : number of points for the circle

    # http://www.physicsforums.com/showthread.php?t=123168
    # P = Rcos(theta))U + Rsin(theta)N x U +c
    # Where u is a unit vector from the centre of the circle
    # to any point on the circumference; R is the radius;
    # n is a unit vector perpendicular to the plane and c is the centre of the circle.

    http://forums.create.msdn.com/forums/p/9551/50048.aspx
    A coworker pointed out a trick to get a vector perpendicular to the normal vector:
    simply swap two of the values, negate one of those, and zero the third.
    So, if I have a normal vector of form Vector3(a, b, c), then one such vector that
    is perpendicular to it is Vector3(b, -a, 0).  Thus, there are six possible vectors
    that are attainable by this method.  The only trouble case is when the normal vector
    contains elements whose values are zero, in which case you have to be a bit careful
    which values you swap and negate.  You just never want to end up with the zero vector.
    """
    N = normal
    U = array([N[1], -N[0], 0], 'f')
    step = PI2/points
    circle = []
    for i in range(points):
        theta = PI2-(step*i)
        P = (r*cos(theta)*U)+(r*sin(theta))*(cross(N,U))+center
        P = normalize(vector(center,P))*r
        P = vecSum(P,center)
        circle.append(P)
    return circle


def quickdist(f,s,sq = False):
    """ works with coordinates/vectors"""
    try:
        d=(f[0]-s[0])**2 + (f[1]-s[1])**2 + (f[2]-s[2])**2
        if sq: return math.sqrt(d)
        else:  return d
    except:
        print("First", f)
        print("Second", s)
        print("WARNING! missing coordinates", sys.exc_info()[1])
        raise Exception
        #return None


def atomsToVector(at1, at2=None, norm=0):

    at1 = atomCoord(at1)
    if at2: at2 = atomCoord(at2)
    return vector(at1, at2, norm=norm)


def vector(p1 , p2 = None, norm = 0): # TODO use Numpy?
    if not p2 is None:
        vec = np.array([p2[0]-p1[0],p2[1]-p1[1],p2[2]-p1[2]],'f')
    else:
        vec = np.array([p1[0], p1[1], p1[2] ], 'f' )
    if norm:
        return normalize(vec)
    else:
        return vec


def norm(A): # TODO use Numpy
    "Return vector norm"
    return np.sqrt(sum(A*A))


def normalize(A): # TODO use Numpy
    "Normalize the Vector"
    return A/norm(A)


def calcPlane(p1, p2, p3):
    # returns the plane containing the 3 input points
    v12 = vector(p1,p2)
    v13 = vector(p3,p2)
    return normalize(np.cross(v12, v13))


def dot(vector1, vector2):  # TODO remove and use Numpy
    dot_product = 0.
    for i in range(0, len(vector1)):
        dot_product += (vector1[i] * vector2[i])
    return dot_product


def vecAngle(v1, v2, rad=1): # TODO remove and use Numpy?
    angle = dot(normalize(v1), normalize(v2))
    if np.array_equal(v1, v2):
        return 0
    try:
        if rad:
            return math.acos(angle)
        else:
            return math.degrees(math.acos(angle))
    except ValueError:
        print("#vecAngle> CHECK TrottNormalization", v1, v2, sys.exc_info()[1])
        return 0


def absoluteAngleDifference(angle1, angle2, rad=1):
    """ https://gamedev.stackexchange.com/questions/4467/comparing-angles-and-working-out-the-difference"""
    ref = 180
    if rad:
        ref = math.radians(180)
    diff = ref - abs( abs(angle1 - angle2) - ref)
    return diff


def vecSum(vec1, vec2): # TODO remove and use Numpy # TODO to be used in the PDBQT+ data!
    return array([vec1[0]+vec2[0], vec1[1]+vec2[1], vec1[2]+vec2[2] ], 'f')


def normValue(v, vmin, vmax, normrange=[0,10]):
    # http://mathforum.org/library/drmath/view/60433.html
    # min = A
    # max = B
    # v   = x
    # y = 1 + (x-A)*(10-1)/(B-A)
    #return  1 + (v-vmin)*(10-1)/(vmax-vmin)
    return  normrange[0] + (v-vmin)*( normrange[1] )/(vmax-vmin)
    #top = (v-vmin)(10-1)
    #down = (vmax-vmin)
    #x =  1 + top/down
    #return x


def normProduct(a, b, mode = 'simple'):
    if mode =='simple': return a*b
    elif mode =='scaled': return (a*b)*(a+b)


def avgVector_untested(vec_list, normalize=False):
    # XXX NOT WORKING!!!
    # http://devmaster.net/forums/topic/5443-average-direction-vector/
    #weight = 1;
    #average = vec[0];
    #for (i = 1; i < n; ++i)
    #{
    #    find angle between average and vec[i];
    #    angle *= weight / (weight + 1);
    #    average = rotate vec[i] towards average by angle;
    #    ++weight;
    #}
    print("avgVector> NOT WORKING!!!! NEVER TESTED")

    weight = 1
    average = vec_list[0]
    for i in range(len(vec_list) - 1):
        angle = vecAngle(average, vec_list[i + 1])
        angle *= weight / (weight + 1)
        average = rotate_around_axis(vec_list[i + 1], m, ax)
        # XXX m?
        # XXX ax?
        weight += 1
    return average


def averageVector(vectorList, norm=True):
    """ """
    vector = np.array([0.,0.,0.], 'f')
    for v in vectorList:
        vector += v
    vector = vector/len(vectorList)
    if norm:
        vector = normalize(vector)
    return vector


def coplanar(plane, coord_list = [], reference = [0., 0., 0.], tolerance = 0.2):
    """ return list of coordinates that are within <tolerance>
        from the plane. If the reference is provided, vectors will be
        calculated with <reference> as origin.

    """
    coplane_list = []
    for c in coord_list:
        pos = vector(reference, c)
        if dot(plane, pos) <= tolerance:
            coplane_list.append(c)
    return coplane_list


def calcRingCentroidNormal(atomCoords):
    """ extract aromatic ring geometric info from a numpy array """
    a1 = atomCoords[0]
    a2 = atomCoords[1]
    a3 = atomCoords[2]

    centroid = averageCoords(atomCoords)
    plane = calcPlane(a1, a2, a3)
    v1 = vector(centroid, a1)
    v2 = vector(centroid, a2)
    normal1 = normalize(np.cross(v1, v2))
    normal2 = normalize(np.cross(v2, v1))
    centroid_norm1 = normalize(vector(centroid, normal1))
    centroid_norm2 = normalize(vector(centroid, normal2))
    return {'centroid':centroid, 'plane':plane, 'normals':[normal1, normal2],
            'centroid_normals':[centroid_norm1, centroid_norm2]}


def gaussian(x, ymax = 1., center=0., spread=0.7, invert=False):
    """ simple gaussian function"""
    if invert:
        invert = -1
    else:
        invert = 1
    return ymax * e **( -((x-center)**2/ (2*spread**2) ) ) * invert


def ellipticGaussian(coord, pseudo, planeVec, centroid=None, dist=None, ellipticity = 1.0,
    g_ymax=1.0, g_center=0.0, g_spread=1.2, g_invert=True):
    """ calculate elliptical gaussian potential based on the angle with the plane
          ____
       .--        __.M
      <       P--'
       `''----|
              |
        ------O-----  plane
    """
    vec = vector(coord, pseudo)
    theta = vecAngle( vec, planeVec)
    if dist is None:
        d = quickdist(coord, pseudo, sq=1)
    else:
        d = dist
    d += (d * ellipticity * (math.fabs(math.cos(theta))))
    ## this code creates a concave cone-shaped ring
    ## d += -(d * ellipticity * (math.fabs(math.cos(theta))))
    gvalue = gaussian(d, ymax = g_ymax, center = g_center,
            spread=g_spread, invert=g_invert)
    return gvalue



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import fnmatch
from setuptools import setup, find_packages


# Path to the directory that contains this setup.py file.
base_dir = os.path.abspath(os.path.dirname(__file__))

def find_files(directory):
    matches = []

    for root, dirnames, filenames in os.walk(directory):
        for filename in fnmatch.filter(filenames, '*'):
            matches.append(os.path.join(root, filename))

    return matches


setup(
    name="meeko",
    version='0.5.1',
    author="Forli Lab",
    author_email="forli@scripps.edu",
    url="https://github.com/ccsb-scripps/meeko",
    description='Python package for preparing small molecule for docking',
    long_description=open(os.path.join(base_dir, 'README.md')).read(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=['numpy>=1.18'],
    python_requires='>=3.5, <3.12',
    license="LGPL-2.1",
    keywords=["molecular modeling", "drug design",
            "docking", "autodock"],
    classifiers=[
        'Environment :: Console',
        'Environment :: Other Environment',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: OS Independent',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: C++',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Software Development :: Libraries'
    ],
    entry_points={
        'console_scripts': [
            'mk_export.py=meeko.cli.mk_export:main',
            'mk_prepare_ligand.py=meeko.cli.mk_prepare_ligand:main',
            'mk_prepare_receptor.py=meeko.cli.mk_prepare_receptor:main'
        ]
    }
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko atom typer
#

import os
import json

import numpy as np

from .utils import utils
from .utils import pdbutils


class AtomTyper:

    defaults_json = """{
        "ATOM_PARAMS": {
            "alkyl glue": [
                {"smarts": "[#1]",                  "atype": "H", "comment": "invisible"},
                {"smarts": "[#1][#7,#8,#9,#15,#16]","atype": "HD"},
                {"smarts": "[#5]",              "atype": "B"},
                {"smarts": "[C]",               "atype": "C"},
                {"smarts": "[c]",               "atype": "A"},
                {"smarts": "[#7]",              "atype": "NA"},
                {"smarts": "[#8]",              "atype": "OA"},
                {"smarts": "[#9]",              "atype": "F"},
                {"smarts": "[#12]",             "atype": "Mg"},
                {"smarts": "[#14]",             "atype": "Si"},
                {"smarts": "[#15]",             "atype": "P"},
                {"smarts": "[#16]",             "atype": "S"},
                {"smarts": "[#17]",             "atype": "Cl"},
                {"smarts": "[#20]",             "atype": "Ca"},
                {"smarts": "[#25]",             "atype": "Mn"},
                {"smarts": "[#26]",             "atype": "Fe"},
                {"smarts": "[#30]",             "atype": "Zn"},
                {"smarts": "[#35]",             "atype": "Br"},
                {"smarts": "[#53]",             "atype": "I"},
                {"smarts": "[#7X3v3][a]",       "atype": "N",  "comment": "pyrrole, aniline"},
                {"smarts": "[#7X3v3][#6X3v4]",  "atype": "N",  "comment": "amide"},
                {"smarts": "[#7+1]",            "atype": "N",  "comment": "ammonium, pyridinium"},
                {"smarts": "[SX2]",             "atype": "SA", "comment": "sulfur acceptor"}
            ]
        }
    }
    """
    def __init__(self, parameters={}, add_parameters=[]):
        self.parameters = json.loads(self.defaults_json)
        for key in parameters:
            self.parameters[key] = json.loads(json.dumps(parameters[key])) # a safe copy
        # add additional parameters
        if len(add_parameters) > 0:
            keys = list(self.parameters["ATOM_PARAMS"].keys())
            if len(keys) != 1:
                msg = "add_parameters is usable only when there is one group of parameters"
                msg += ", but there are %d groups: %s" % (len(keys), str(keys))
                raise RuntimeError(msg)
            key = keys[0]
            self.parameters['ATOM_PARAMS'][key].extend(add_parameters)

    def __call__(self, setup):
        self._type_atoms(setup)
        if 'OFFATOMS' in self.parameters:
            cached_offatoms = self._cache_offatoms(setup)
            coords = [x for x in setup.coord.values()]
            self._set_offatoms(setup, cached_offatoms, coords)
        return

    def _type_atoms(self, setup):
        parsmar = self.parameters['ATOM_PARAMS']
        # ensure every "atompar" is defined in a single "smartsgroup"
        ensure = {}
        # go over all "smartsgroup"s
        for smartsgroup in parsmar:
            if smartsgroup == 'comment': continue
            for line in parsmar[smartsgroup]: # line is a dict, e.g. {"smarts": "[#1][#7,#8,#9,#15,#16]","atype": "HD"}
                smarts = str(line['smarts'])
                if 'atype' not in line: continue
                # get indices of the atoms in the smarts to which the parameters will be assigned
                idxs = [0] # by default, the first atom in the smarts gets parameterized
                if 'IDX' in line:
                    idxs = [i - 1 for i in line['IDX']] # convert from 1- to 0-indexing
                # match SMARTS
                hits = setup.find_pattern(smarts)
                atompar = 'atype' # we care only about 'atype', for now, but may want to extend
                atom_type = line[atompar]
                # keep track of every "smartsgroup" that modified "atompar"
                ensure.setdefault(atompar, [])
                ensure[atompar].append(smartsgroup)
                # Each "hit" is a tuple of atom indeces that matched the smarts
                # The length of each "hit" is the number of atoms in the smarts
                for hit in hits:
                    # Multiple atoms may be targeted by a single smarts:
                    # For example: both oxygens in NO2 are parameterized by a single smarts pattern.
                    # "idxs" are 1-indeces of atoms in the smarts to which parameters are to be assigned.
                    for idx in idxs:
                        setup.set_atom_type(hit[idx], atom_type) # overrides previous calls
        # guarantee that each atompar is exclusive of a single group
        for atompar in ensure:
            if len(set(ensure[atompar])) > 1:
                msg = 'WARNING: %s is modified in multiple smartsgroups: %s' % (atompar, set(ensure[atompar]))
                print(msg)
        return


    def _cache_offatoms(self, setup):
        """ precalculate off-site atoms """
        parsmar = self.parameters['OFFATOMS']
        cached_offatoms = {}
        n_offatoms = 0
        # each parent atom can only be matched once in each smartsgroup
        for smartsgroup in parsmar:
            if smartsgroup == "comment": continue
            tmp = {}
            for line in parsmar[smartsgroup]:
                # SMARTS
                smarts = str(line['smarts'])
                hits = setup.find_pattern(smarts)
                # atom indexes in smarts string
                smarts_idxs = [0]
                if 'IDX' in line:
                    smarts_idxs = [i - 1 for i in line['IDX']]
                for smarts_idx in smarts_idxs:
                    for hit in hits:
                        parent_idx = hit[smarts_idx]
                        tmp.setdefault(parent_idx, []) # TODO tmp[parent_idx] = [], yeah?
                        for offatom in line['OFFATOMS']:
                            # set defaults
                            tmp[parent_idx].append(
                                {'offatom': {'distance': 1.0,
                                             'x90': False,
                                             'phi': 0.0,
                                             'theta': 0.0,
                                             'z': [],
                                             'x': []},
                                 'atom_params': {}
                                })
                            for key in offatom:
                                if key in ['distance', 'x90']:
                                    tmp[parent_idx][-1]['offatom'][key] = offatom[key]
                                # replace SMARTS indexes by the atomic index
                                elif key in ['z', 'x']:
                                    for i in offatom[key]:
                                        idx = hit[i - 1]
                                        tmp[parent_idx][-1]['offatom'][key].append(idx)
                                # convert degrees to radians
                                elif key in ['theta', 'phi']:
                                    tmp[parent_idx][-1]['offatom'][key] = np.radians(offatom[key])
                                # ignore comments
                                elif key in ['comment']:
                                    pass
                                elif key == 'atype':
                                    tmp[parent_idx][-1]['atom_params'][key] = offatom[key]
                                else:
                                    pass
            for parent_idx in tmp:
                for offatom_dict in tmp[parent_idx]:
                    #print '1-> ', self.atom_params['q'], len(self.coords)
                    atom_params = offatom_dict['atom_params']
                    offatom = offatom_dict['offatom']
                    atomgeom = AtomicGeometry(parent_idx,
                                              neigh=offatom['z'],
                                              xneigh=offatom['x'],
                                              x90=offatom['x90'])
                    args = (atom_params['atype'],
                            offatom['distance'],
                            offatom['theta'],
                            offatom['phi'])
                    # number of coordinates (before adding new offatom)
                    cached_offatoms[n_offatoms] = (atomgeom, args)
                    n_offatoms += 1
        return cached_offatoms

    def _set_offatoms(self, setup, cached_offatoms, coords):
        """add cached offatoms"""
        for k, (atomgeom, args) in cached_offatoms.items():
            (atom_type, dist, theta, phi) = args
            offatom_coords = atomgeom.calc_point(dist, theta, phi, coords)
            tmp = setup.get_pdbinfo(atomgeom.parent+1)
            pdbinfo = pdbutils.PDBAtomInfo('G', tmp.resName, tmp.resNum, tmp.chain)
            pseudo_atom = {
                    'coord': offatom_coords,
                    'anchor_list': [atomgeom.parent],
                    'charge': 0.0,
                    'pdbinfo': pdbinfo,
                    'atom_type': atom_type,
                    'bond_type': 0,
                    'rotatable': False
                    }
            setup.add_pseudo(**pseudo_atom)
        return

class AtomicGeometry():
    """generate reference frames and add extra sites"""

    PLANAR_TOL = 0.1 # angstroms, length of neighbour vecs for Z axis

    def __init__(self, parent, neigh, xneigh=[], x90=False):
        """arguments are indices of atoms"""

        # real atom hosting extra sites
        if type(parent) != int:
            raise RuntimeError('parent must be int')
        self.parent = parent

        # list of bonded atoms (used to define z-axis)
        self.neigh = []
        for i in neigh:
            if type(i) != int:
                raise RuntimeError('neigh indices must be int')
            self.neigh.append(i)

        # list of atoms that
        self.xneigh = []
        for i in xneigh:
            if type(i) != int:
                raise RuntimeError('xneigh indices must be int')
            self.xneigh.append(i)

        self.calc_x = len(self.xneigh) > 0
        self.x90 = x90 # y axis becomes x axis (useful to rotate in-plane by 90 deg)

    def calc_point(self, distance, theta, phi, coords):
        """return coordinates of point specified in spherical coordinates"""

        z = self._calc_z(coords)

        # return pt aligned with z-axis
        if phi == 0.:
            return z * distance + np.array(coords[self.parent])

        # need x-vec if phi != 0
        elif self.calc_x == False:
            raise RuntimeError('phi must be zero if X undefined')

        else:
            x = self._calc_x(coords)
            if self.x90:
                x = np.cross(self.z, x)
            y = np.cross(z, x)
            pt = z * distance
            pt = self._rot3D(pt, y, phi)
            pt = self._rot3D(pt, z, theta)
            pt += np.array(coords[self.parent])
            return pt

    def _calc_z(self, coords):
        """ maximize distance from neigh """
        z = np.zeros(3)
        cumsum = np.zeros(3)
        for i in self.neigh:
            v = np.array(coords[self.parent]) - np.array(coords[i])
            cumsum += v
            z += self.normalized(v)
        z = self.normalized(z)
        if np.sum(cumsum**2) < self.PLANAR_TOL**2:
            raise RuntimeError('Refusing to place Z axis on planar atom')
        return z

    def _calc_x(self, coords):
        x = np.zeros(3)
        for i in self.xneigh:
            v = np.array(coords[self.parent]) - np.array(coords[i])
            x += self.normalized(v)
        x = self.normalized(x)
        return x

    def _rot3D(self, pt, ax, rad):
        """
            Rotate point:
            pt = (x,y,z) coordinates to be rotated
            ax = vector around wich rotation is performed
            rad = rotate by "rad" radians
        """
        # If axis has len=0, rotate by 0.0 rad on any axis
        # Make sure ax has unitary length
        len_ax = (ax[0]**2 + ax[1]**2 + ax[2]**2)**0.5
        if len_ax == 0.:
            u, v, w = (1, 0, 0)
            rad = 0.0
        else:
            u, v, w = [i/len_ax for i in ax]
        x, y, z = pt
        ux, uy, uz = u*x, u*y, u*z
        vx, vy, vz = v*x, v*y, v*z
        wx, wy, wz = w*x, w*y, w*z
        sa=np.sin(rad)
        ca=np.cos(rad)
        p0 =(u*(ux+vy+wz)+(x*(v*v+w*w)-u*(vy+wz))*ca+(-wy+vz)*sa)
        p1=(v*(ux+vy+wz)+(y*(u*u+w*w)-v*(ux+wz))*ca+(wx-uz)*sa)
        p2=(w*(ux+vy+wz)+(z*(u*u+v*v)-w*(ux+vy))*ca+(-vx+uy)*sa)
        return (p0, p1, p2)


    def normalized(self, vec):
        l = sum([x**2 for x in vec])**0.5
        if type(vec) == list:
            return [x/l for x in vec]
        else:
            # should be np.array
            return vec / l
